from typing import List
from depccg.utils import denormalize
from depccg.tree import Tree


def _resolve_dependencies(tree: Tree) -> List[int]:
    results = []

    def rec(node: Tree) -> int:
        if node.is_leaf:
            index = len(results)
            results.append(-1)
            return index
        else:
            if node.is_unary:
                return rec(node.child)
            else:
                left_head = rec(node.left_child)
                right_head = rec(node.right_child)
                if node.head_is_left:
                    results[right_head] = left_head
                    return left_head
                else:
                    results[left_head] = right_head
                    return right_head

    rec(tree)

    assert len(
        [dependency for dependency in results if dependency == -1]
    ) == 1

    return results


def conll_of(tree: Tree) -> str:
    """CoNLL-like format string where dependency relations are constructed by
    nodes' head_is_left property.

    Args:
        tree (Tree): tree object
        tokens (Optional[List[Token]], optional): list of token objects. Defaults to None.

    Returns:
        str: tree string in the CoNLL format
    """

    stack = []
    counter = 1

    def rec(node):
        nonlocal stack, counter
        if node.is_leaf:
            cat = node.cat
            word = denormalize(node.word)
            token = node.token
            lemma = token.get('lemma', '_')
            pos = token.get('pos', '_')
            stack.append(f'(<L {cat} {pos} {pos} {word} {cat}>)')
            subtree = ' '.join(stack)

            line = '\t'.join(
                (
                    str(counter),
                    word,
                    lemma,
                    pos,
                    pos,
                    '_',
                    str(dependencies[counter - 1] + 1),
                    str(cat),
                    '_',
                    subtree
                )
            )

            stack = []
            counter += 1
            return line
        else:
            cat = node.cat
            num_children = len(node.children)
            head_is_left = 0 if node.head_is_left else 1
            stack.append(f'(<T {cat} {head_is_left} {num_children}>')
            children = '\n'.join(rec(child) for child in node.children) + ' )'
            return children

    dependencies = _resolve_dependencies(tree)
    return rec(tree)



# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import re
import simplejson

from lxml import etree
from nltk.sem.logic import ConstantExpression

from .logic_parser import lexpr
from .normalization import normalize_token
import depccg.semantics.ccg2lambda.semantic_index as semantic_index

def build_ccg_tree(ccg_xml, root_id=None):
    """
    This function re-arranges the nodes of the XML CCG tree to have
    a tree structure. It will be useful to traverse the tree.
    """
    if ccg_xml == None or len(ccg_xml) == 0:
        return None
    if root_id == None:
        root_id = ccg_xml.get('root')
    root_span = copy.deepcopy(semantic_index.find_node_by_id(root_id, ccg_xml))
    if 'child' not in root_span.attrib:
        return root_span
    children_id = root_span.get('child').split()
    for child_id in children_id:
        child_node = build_ccg_tree(ccg_xml, child_id)
        if child_node != None:
            root_span.append(child_node)
    return root_span

def normalize_tokens(tokens):
    """
    In our format of XML trees, tokens have their own tree,
    which is separated from the syntactic structure. These
    tokens may need some processing for normalization, such
    as prefixing them with an underscore "_", or copying
    into their base form the surface form when the base form
    is absent (base="*"). For now, only the copy to base forms
    from surface forms is implemented:
    """
    for token in tokens:
        if token.get('base', None) == '*':
            token.set('base', token.get('surf', '*'))
        if 'base' in token.attrib and not token.get('base', '').startswith('_'):
            base = token.get('base', '')
            base_normalized = normalize_token(base)
            token.set('base', base_normalized)
        if 'surf' in token.attrib and not token.get('surf', '').startswith('_'):
            surf = token.get('surf', '')
            surf_normalized = normalize_token(surf)
            token.set('surf', surf_normalized)
    return tokens

def assign_semantics_to_ccg(ccg_xml, semantic_index, tree_index=1):
    """
    This is the key function. It builds first an XML tree structure with
    the CCG tree, and then assigns semantics (lambda expressions) to each node
    in post-order (first assigns semantics to children, and then to node).
    It returns a CCG lxml tree structure with a new 'sem' field that
    contains the semantics at each node.
    """
    # If the use of gold trees is requested, we get the gold tree index
    # from the XML attribute 'gold_tree'. Note that in xpath, lists are
    # one-indexed arrays.
    # from pudb import set_trace; set_trace()
    ccg_flat_trees = ccg_xml.xpath('./ccg[{0}]'.format(tree_index))
    if not ccg_flat_trees:
        raise ValueError(
            'No CCG tree found in:\n{0}'.format(
            etree.tostring(
                ccg_xml,
                encoding='utf-8',
                pretty_print=True).decode('utf-8')))
    ccg_flat_tree = copy.deepcopy(ccg_flat_trees[0])
    #   ccg_xml.xpath('./ccg[{0}]'.format(tree_index))[0])
    ccg_tree = build_ccg_tree(ccg_flat_tree)
    tokens = copy.deepcopy(ccg_xml.find('.//tokens'))
    tokens = normalize_tokens(tokens)
    assign_semantics(ccg_tree, semantic_index, tokens)
    return ccg_tree

def is_forward_operation(ccg_tree):
    rule = ccg_tree.get('rule')
    if 'gt' in rule or '>' in rule or (len(rule) > 0 and rule[0] == 'f'):
        return True
    return False

def get_combination_op(ccg_tree):
    """
    As of today, CCG trees as produce by transccg does not signal
    function combination. This code assumes that transccg would insert
    the character 'B' somewhere in the rule, to easy identification of operation.
    """
    rule = ccg_tree.get('rule')
    if 'B' in rule or (len(rule) > 1 and (rule[1] not in ['a', 'p'])):
        return 'function_combination'
    return 'function_application'

def get_num_args(ccg_tree):
    rule = ccg_tree.get('rule')
    num_args = 1
    if '2' in rule:
        num_args = 2
    elif '3' in rule:
        num_args = 3
    return num_args

def type_raise(function, order = 1):
    """
    Produce a higher order function based on "function". The argument "order"
    indicates the number of desired arguments of the new function.
    """
    assert order >= 0, 'The order of the type-raising should be >= 0'
    if isinstance(function, ConstantExpression):
        type_raiser = lexpr(r'\P X.P(X)')
        type_raised_function = type_raiser(function).simplify()
    else:
        if order == 1:
            type_raiser = lexpr(r'\P0 P1 X0.P0(P1(X0))')
        elif order == 2:
            type_raiser = lexpr(r'\P0 P1 X0 X1.P0(P1(X0, X1))')
        elif order == 3:
            type_raiser = lexpr(r'\P0 P1 X0 X1 X2.P0(P1(X0, X1, X2))')
        else:
            assert False, 'Type-raising at order > 3 is not supported'
        type_raised_function = type_raiser(function).simplify()
    return type_raised_function

def combine_children_exprs(ccg_tree, tokens, semantic_index):
    """
    Perform forward/backward function application/combination.
    """
    assert len(ccg_tree) >= 2, \
      'There should be at least two children to combine expressions: {0}'\
      .format(ccg_tree)
    # Assign coq types.
    coq_types_left  = ccg_tree[0].attrib.get('coq_type', "")
    coq_types_right = ccg_tree[1].attrib.get('coq_type', "")
    if coq_types_left and coq_types_right:
        coq_types = coq_types_left + ' ||| ' + coq_types_right
    elif coq_types_left:
        coq_types = coq_types_left 
    else:
        coq_types = coq_types_right
    ccg_tree.set('coq_type', coq_types)
    semantics = semantic_index.get_semantic_representation(ccg_tree, tokens)
    if semantics:
        ccg_tree.set('sem', str(semantics))
        return None
    # Back-off mechanism in case no semantic templates are available:
    if is_forward_operation(ccg_tree):
        function_index, argument_index = 0, 1
    else:
        function_index, argument_index = 1, 0
    function = lexpr(ccg_tree[function_index].attrib['sem'])
    argument = lexpr(ccg_tree[argument_index].attrib['sem'])
    combination_operation = get_combination_op(ccg_tree)
    if combination_operation == 'function_application':
        evaluation = function(argument).simplify()
    elif combination_operation == 'function_combination':
        num_arguments = get_num_args(ccg_tree)
        type_raised_function = type_raise(function, num_arguments)
        evaluation = type_raised_function(argument).simplify()
    else:
        assert False, 'This node should be a function application or combination'
    ccg_tree.set('sem', str(evaluation))
    return None

def assign_semantics(ccg_tree, semantic_index, tokens):
    """
    Visit recursively the CCG tree in depth-first order, assigning lambda expressions
    (semantics) to each node.
    """
    category = ccg_tree.attrib['category']
    if len(ccg_tree) == 0:
        semantics = semantic_index.get_semantic_representation(ccg_tree, tokens)
        ccg_tree.set('sem', str(semantics))
        return
    if len(ccg_tree) == 1:
        assign_semantics(ccg_tree[0], semantic_index, tokens)
        semantics = semantic_index.get_semantic_representation(ccg_tree, tokens)
        ccg_tree.set('sem', str(semantics))
        return
    for child in ccg_tree:
        assign_semantics(child, semantic_index, tokens)
    combine_children_exprs(ccg_tree, tokens, semantic_index)
    return



from typing import Optional, List, Dict, Set, TypeVar, Tuple
from depccg.cat import Category
from depccg.unification import Unification
from depccg.types import Combinator, CombinatorResult

X = TypeVar('X')
Pair = Tuple[X, X]


def _is_modifier(x: Category) -> bool:
    return x.is_functor and x.left == x.right


def forward_application(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "b")
    if uni(x, y):
        result = y if _is_modifier(x) else uni['a']
        return CombinatorResult(
            cat=result,
            op_string="fa",
            op_symbol=">",
            head_is_left=False,
        )
    return None


def backward_application(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("b", "a\\b")
    if uni(x, y):
        result = x if _is_modifier(y) else uni['a']
        return CombinatorResult(
            cat=result,
            op_string="ba",
            op_symbol="<",
            head_is_left=False,
        )
    return None


def forward_composition(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "b/c")
    if uni(x, y):
        result = y if _is_modifier(x) else uni['a'] / uni['c']
        return CombinatorResult(
            cat=result,
            op_string="fc",
            op_symbol=">B",
            head_is_left=False,
        )
    return None


def generalized_backward_composition1(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("b\\c", "a\\b")
    if uni(x, y):
        result = x if _is_modifier(y) else uni['a'] | uni['c']
        return CombinatorResult(
            cat=result,
            op_string="bx",
            op_symbol="<B1",
            head_is_left=False,
        )
    return None


def generalized_backward_composition2(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("(b\\c)|d", "a\\b")
    if uni(x, y):
        result = x if _is_modifier(y) else x.functor(
            uni['a'] | uni['c'], uni['d'])
        return CombinatorResult(
            cat=result,
            op_string="bx",
            op_symbol="<B2",
            head_is_left=False,
        )
    return None


def generalized_backward_composition3(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("((b\\c)|d)|e", "a\\b")
    if uni(x, y):
        result = x if _is_modifier(y) else x.functor(
            x.left.functor(uni['a'] | uni['c'], uni['d']), uni['e']
        )
        return CombinatorResult(
            cat=result,
            op_string="bx",
            op_symbol="<B3",
            head_is_left=False,
        )
    return None


def generalized_backward_composition4(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("(((b\\c)|d)|e)|f", "a\\b")
    if uni(x, y):
        result = x if _is_modifier(y) else x.functor(
            x.left.functor(
                x.left.left.functor(uni['a'] | uni['c'], uni['d']),
                uni['e']
            ),
            uni['f']
        )
        return CombinatorResult(
            cat=result,
            op_string="bx",
            op_symbol="<B4",
            head_is_left=False,
        )
    return None


def generalized_forward_composition1(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "b\\c")
    if uni(x, y):
        result = y if _is_modifier(x) else uni['a'] / uni['c']
        return CombinatorResult(
            cat=result,
            op_string="fx",
            op_symbol=">Bx1",
            head_is_left=False,
        )
    return None


def generalized_forward_composition2(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "(b\\c)|d")
    if uni(x, y):
        result = y if _is_modifier(x) else y.functor(
            uni['a'] | uni['c'], uni['d'])
        return CombinatorResult(
            cat=result,
            op_string="fx",
            op_symbol=">Bx2",
            head_is_left=False,
        )
    return None


def generalized_forward_composition3(x: Category, y: Category) -> Optional[CombinatorResult]:
    uni = Unification("a/b", "((b\\c)|d)|e")
    if uni(x, y):
        result = y if _is_modifier(x) else y.functor(y.left.functor(
            uni['a'] | uni['c'], uni['d']), uni['e']
        )
        return CombinatorResult(
            cat=result,
            op_string="fx",
            op_symbol=">Bx3",
            head_is_left=False,
        )
    return None


_possible_root_categories = [
    Category.parse("NP[case=nc,mod=nm,fin=f]"),
    Category.parse("NP[case=nc,mod=nm,fin=t]"),
    Category.parse("S[mod=nm,form=attr,fin=t]"),
    Category.parse("S[mod=nm,form=base,fin=f]"),
    Category.parse("S[mod=nm,form=base,fin=t]"),
    Category.parse("S[mod=nm,form=cont,fin=f]"),
    Category.parse("S[mod=nm,form=cont,fin=t]"),
    Category.parse("S[mod=nm,form=da,fin=f]"),
    Category.parse("S[mod=nm,form=da,fin=t]"),
    Category.parse("S[mod=nm,form=hyp,fin=t]"),
    Category.parse("S[mod=nm,form=imp,fin=f]"),
    Category.parse("S[mod=nm,form=imp,fin=t]"),
    Category.parse("S[mod=nm,form=r,fin=t]"),
    Category.parse("S[mod=nm,form=s,fin=t]"),
    Category.parse("S[mod=nm,form=stem,fin=f]"),
    Category.parse("S[mod=nm,form=stem,fin=t]")
]


def conjoin(x: Category, y: Category) -> Optional[CombinatorResult]:
    if (
        x in _possible_root_categories
        and y in _possible_root_categories
    ):
        # if x == y and x in _possible_root_categories:
        result = y
        return CombinatorResult(
            cat=result,
            op_string="other",
            op_symbol="SSEQ",
            head_is_left=False,
        )


combinators: List[Combinator] = [
    forward_application,
    backward_application,
    forward_composition,
    generalized_backward_composition1,
    generalized_backward_composition2,
    generalized_backward_composition3,
    generalized_backward_composition4,
    generalized_forward_composition1,
    generalized_forward_composition2,
    generalized_forward_composition3,
    conjoin,
]


def apply_binary_rules(
    x: Category,
    y: Category,
    seen_rules: Optional[Set[Pair[Category]]] = None,
) -> List[CombinatorResult]:
    key = (x, y)
    results = []
    if seen_rules is None or key in seen_rules:
        for combinator in combinators:
            result = combinator(*key)
            if result is not None:
                results.append(result)

    return results


def _unary_rule_symbol(x: Category) -> str:

    features = set(x.arg(0).feature.items())
    if ('mod', 'adn') in features:
        if x.clear_features == 'S':
            return 'ADNext'
        return 'ADNint'
    elif ('mod', 'adv') in features:
        if x.clear_features == 'S\\NP':
            return 'ADV1'
        elif x.clear_features == '(S\\NP)\\NP':
            return 'ADV2'
        return 'ADV0'
    return 'OTHER'


def apply_unary_rules(
    x: Category,
    unary_rules: Dict[Category, List[Category]]
) -> List[CombinatorResult]:
    if x not in unary_rules:
        return []
    results = []
    for result in unary_rules[x]:
        op_string = _unary_rule_symbol(x)
        results.append(
            CombinatorResult(
                cat=result,
                op_string=op_string,
                op_symbol=op_string,
                head_is_left=True,
            )
        )
    return results




import argparse
from depccg.tools.reader import read_trees_guess_extension
from depccg.printer import print_
from depccg.annotator import japanese_annotator
from depccg.instance_models import SEMANTIC_TEMPLATES
from depccg.types import ScoringResult
import logging

logger = logging.getLogger(__name__)

LANG = 'ja'

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'PATH',
        help='path to either of *.auto, *.xml, *.jigg.xml, *.ptb')
    parser.add_argument(
        '--annotator',
        default='janome',
        choices=japanese_annotator.keys(),
        help='annotate POS, named entity, and lemmas using this library')
    parser.add_argument(
        '-f',
        '--format',
        default='jigg_xml',
        choices=[
            'prolog', 'jigg_xml',
            'jigg_xml_ccg2lambda', 'json', 'deriv'],
        help='output format')
    parser.add_argument(
        '--semantic-templates',
        help='semantic templates used in "ccg2lambda" format output')
    args = parser.parse_args()

    trees = [
        [ScoringResult(tree, 0)]
        for _, _, tree in read_trees_guess_extension(args.PATH, lang='ja')
    ]
    semantic_templates = args.semantic_templates or SEMANTIC_TEMPLATES[LANG]
    print_(
        trees,
        format=args.format,
        semantic_templates=semantic_templates
    )




import json


class Param:
    def __init__(self, obj):
        self.__dict__["ps"] = {"model": obj.__class__.__name__}
        self.__dict__["obj"] = obj

    def __setattr__(self, key, value):
        self.__dict__["ps"][key] = value
        self.__dict__["obj"].__dict__[key] = value

    def dump(self, out):
        json.dump(self.ps, open(out, "w"))

    @staticmethod
    def load(obj, paramfile):
        p = Param(obj)
        params = json.load(open(paramfile))
        p.__dict__["ps"].update(params)
        obj.__dict__.update(params)
        return p



# -*- coding: utf-8 -*-

from nltk.sem.logic import *
import unicodedata

from nltk.internals import Counter
from .logic_parser import lexpr

_counter = Counter()

# Term t ::=
#   x,y,z         <IndividualVariableExpression>
#   e             <EventVariableExpression>
#   f(t)          <ApplicationExpression> where f in {Subj,Obj,Dat,...}
#   john          <ConstantExpression>
#   \x.student(x) <LambdaExpression>

# Formula A ::=
#   F(t)         <ApplicationExpression>
#   F(t,u)       <ApplicationExpression>
#   t = u        <EqualityExpression>
#   A & B        <AndExpression>
#   A | B        <OrExpression>
#   A -> B       <ImpExpression>
#   -A           <NegatedExpression>
#   exists x.A   <ExistsExpression>
#   forall x.A   <AllExpression>
#   True         <ConstantExpression>

def get_atomic_formulas(expression):
    if isinstance(expression, ApplicationExpression):
        return set([expression])
    elif isinstance(expression, EqualityExpression):
        return set([expression])
    elif isinstance(expression, IndividualVariableExpression):
        return set([expression])
    elif isinstance(expression, EventVariableExpression):
        return set([expression])
    elif isinstance(expression, FunctionVariableExpression):
        return set([expression])
    elif isinstance(expression, ConstantExpression):
        return set([expression])
    else:
        return expression.visit(get_atomic_formulas,
                       lambda parts: reduce(operator.or_, parts, set()))

def new_variable(var):
    var = VariableExpression(var)
    # isinstance(var,EventVariableExpression) must come first
    if isinstance(var, EventVariableExpression):
        prefix = 'e0'
    elif isinstance(var, IndividualVariableExpression):
        prefix = 'x0'
    elif isinstance(var, FunctionVariableExpression):
        prefix = 'F0'
    else:
        prefix = 'z0'
    v = Variable("%s%s" % (prefix, _counter.get()))
    return v

true_preds = ['True', 'TrueP']

def remove_true(expression):
    # Remove True and TrueP
    if isinstance(expression, ApplicationExpression):
        function = remove_true(expression.function)
        argument = remove_true(expression.argument)
        expr = ApplicationExpression(function, argument)
    elif isinstance(expression, EqualityExpression):
        left = remove_true(expression.first)
        right = remove_true(expression.second)
        expr = EqualityExpression(left, right)
    elif isinstance(expression, AndExpression):
        # True & A <=> A & True <=> A
        left = expression.first
        right = expression.second
        left_str = str(left)
        right_str = str(right)
        if left_str in true_preds:
            expr = remove_true(right)
        elif right_str in true_preds:
            expr = remove_true(left)
        else:
            left = remove_true(left)
            right = remove_true(right)
            expr = AndExpression(left, right)
    elif isinstance(expression, OrExpression):
        # True or A <=> A or True <=> True
        left = expression.first
        right = expression.second
        left_str = str(left)
        right_str = str(right)
        if left_str in true_preds:
            expr = remove_true(left)
        elif right_str in true_preds:
            expr = remove_true(right)
        else:
            left = remove_true(left)
            right = remove_true(right)
            expr = OrExpression(left, right)
    elif isinstance(expression, ImpExpression):
        # True -> A <=> A
        left = expression.first
        right = expression.second
        left_str = str(left)
        if left_str in true_preds:
            expr = remove_true(right)
        else:
            left = remove_true(expression.first)
            right = remove_true(expression.second)
            expr = ImpExpression(left, right)
    elif isinstance(expression, NegatedExpression):
        term = remove_true(expression.term)
        expr = NegatedExpression(term)
    elif isinstance(expression, ExistsExpression):
        variable = expression.variable
        term = expression.term
        newvar = new_variable(variable)
        newvar_expr = VariableExpression(newvar)
        term = term.replace(variable, newvar_expr)
        term = remove_true(term)
        expr = ExistsExpression(newvar, term)
    elif isinstance(expression, AllExpression):
        variable = expression.variable
        term = expression.term
        newvar = new_variable(variable)
        newvar_expr = VariableExpression(newvar)
        term = term.replace(variable, newvar_expr)
        term = remove_true(term)
        expr = AllExpression(newvar, term)
    elif isinstance(expression, LambdaExpression):
        variable = expression.variable
        term = expression.term
        newvar = new_variable(variable)
        newvar_expr = VariableExpression(newvar)
        term = term.replace(variable, newvar_expr)
        term = remove_true(term)
        expr = LambdaExpression(newvar, term)
    # elif isinstance(expression, IndividualVariableExpression):
    #     expr = expression
    # elif isinstance(expression, EventVariableExpression):
    #     expr = expression
    # elif isinstance(expression, FunctionVariableExpression):
    #     expr = expression
    # elif isinstance(expression, ConstantExpression):
    #     expr = expression
    else:
        expr = expression
    return expr


def rename_variable(expression):
    # Rename bound variables so that no variable with the same name is bound
    # by two different quantifiers in different parts of a formula
    if isinstance(expression, ApplicationExpression):
        function = rename_variable(expression.function)
        argument = rename_variable(expression.argument)
        expr = ApplicationExpression(function, argument)
    elif isinstance(expression, EqualityExpression):
        left = rename_variable(expression.first)
        right = rename_variable(expression.second)
        expr = EqualityExpression(left, right)
    elif isinstance(expression, AndExpression):
        left = rename_variable(expression.first)
        right = rename_variable(expression.second)
        expr = AndExpression(left, right)
    elif isinstance(expression, OrExpression):
        left = rename_variable(expression.first)
        right = rename_variable(expression.second)
        expr = OrExpression(left, right)
    elif isinstance(expression, ImpExpression):
        left = rename_variable(expression.first)
        right = rename_variable(expression.second)
        expr = ImpExpression(left, right)
    elif isinstance(expression, NegatedExpression):
        term = rename_variable(expression.term)
        expr = NegatedExpression(term)
    elif isinstance(expression, ExistsExpression):
        variable = expression.variable
        term = expression.term
        newvar = new_variable(variable)
        newvar_expr = VariableExpression(newvar)
        term = term.replace(variable, newvar_expr)
        term = rename_variable(term)
        expr = ExistsExpression(newvar, term)
    elif isinstance(expression, AllExpression):
        variable = expression.variable
        term = expression.term
        newvar = new_variable(variable)
        newvar_expr = VariableExpression(newvar)
        term = term.replace(variable, newvar_expr)
        term = rename_variable(term)
        expr = AllExpression(newvar, term)
    elif isinstance(expression, LambdaExpression):
        variable = expression.variable
        term = expression.term
        newvar = new_variable(variable)
        newvar_expr = VariableExpression(newvar)
        term = term.replace(variable, newvar_expr)
        term = rename_variable(term)
        expr = LambdaExpression(newvar, term)
    # elif isinstance(expression, IndividualVariableExpression):
    #     expr = expression
    # elif isinstance(expression, EventVariableExpression):
    #     expr = expression
    # elif isinstance(expression, FunctionVariableExpression):
    #     expr = expression
    # elif isinstance(expression, ConstantExpression):
    #     expr = expression
    else:
        expr = expression
    return expr

def convert_to_prenex(expression):
    # Convert a formula to one where all existential quantifers come first.
    expression = remove_true(expression)
    expression = rename_variable(expression)
    prenex_form = prenex_expr(expression)
    return prenex_form

def prenex_expr(expression):
    if isinstance(expression, ApplicationExpression):
        expr = prenex_application_expr(expression)
    elif isinstance(expression, EqualityExpression):
        expr = prenex_equality_expr(expression)
    elif isinstance(expression, AndExpression):
        expr = prenex_and_expr(expression)
    elif isinstance(expression, OrExpression):
        expr = prenex_or_expr(expression)
    elif isinstance(expression, ImpExpression):
        expr = prenex_imp_expr(expression)
    elif isinstance(expression, NegatedExpression):
        expr = prenex_not_expr(expression)
    elif isinstance(expression, ExistsExpression):
        expr = prenex_exists_expr(expression)
    elif isinstance(expression, AllExpression):
        expr = prenex_all_expr(expression)
    elif isinstance(expression, LambdaExpression):
        expr = prenex_lambda_expr(expression)
    elif isinstance(expression, IndividualVariableExpression):
        expr = expression
    elif isinstance(expression, EventVariableExpression):
        expr = expression
    elif isinstance(expression, ConstantExpression):
        lexstr = normalize_symbols('%s' % expression)
        expr = ConstantExpression(Variable(lexstr))
    # elif isinstance(expression, Variable):
    #     expr = expression
    else:
        expr = expression
    return expr

def prenex_application_expr(expression):
    function = prenex_expr(expression.function)
    argument = prenex_expr(expression.argument)
    expr = ApplicationExpression(function, argument)
    return expr

def prenex_equality_expr(expression):
    left = prenex_expr(expression.first)
    right = prenex_expr(expression.second)
    expr = EqualityExpression(left, right)
    return expr

def prenex_and_expr(expression):
    left = prenex_expr(expression.first)
    right = prenex_expr(expression.second)
    # [(exists x. L) & R] = exists x. [L & R]
    if isinstance(left, ExistsExpression):
        left_variable = left.variable
        left_term = left.term
        if not left_variable in right.free():
            body = prenex_expr(AndExpression(left_term, right))
            expr = ExistsExpression(left_variable, body)
        else:
            newvar = unique_variable()
            var = VariableExpression(newvar)
            left_term = left_term.replace(left_variable, var)
            body = prenex_expr(AndExpression(left_term, right))
            expr = ExistsExpression(newvar, body)
    # [L & (exists x. R)] = exists x. [L & R]
    elif isinstance(right, ExistsExpression):
        right_variable = right.variable
        right_term = right.term
        if not right_variable in left.free():
            body = prenex_expr(AndExpression(left, right_term))
            expr = ExistsExpression(right_variable, body)
        else:
            newvar = unique_variable()
            var = VariableExpression(newvar)
            right_term = right_term.replace(right_variable, var)
            body = prenex_expr(AndExpression(left, right_term))
            expr = ExistsExpression(newvar, body)
    else:
        expr = AndExpression(left, right)
    return expr

def prenex_or_expr(expression):
    left = prenex_expr(expression.first)
    right = prenex_expr(expression.second)
    expr = OrExpression(left, right)
    return expr

def prenex_imp_expr(expression):
    left = prenex_expr(expression.first)
    right = prenex_expr(expression.second)
    expr = ImpExpression(left, right)
    return expr

def prenex_not_expr(expression):
    term = prenex_expr(expression.term)
    expr = NegatedExpression(term)
    return expr

def prenex_exists_expr(expression):
    variable = expression.variable
    term = prenex_expr(expression.term)
    expr = ExistsExpression(variable, term)
    return expr

def prenex_all_expr(expression):
    variable = expression.variable
    term = prenex_expr(expression.term)
    expr = AllExpression(variable, term)
    return expr

def prenex_lambda_expr(expression):
    variable = expression.variable
    term = prenex_expr(expression.term)
    expr = LambdaExpression(variable, term)
    return expr

def normalize_symbols(expression):
  expression = expression.replace("’","").\
               replace("_","").\
               replace("（","BracketLeft").\
               replace("）","BracketRight")
  expression = unicodedata.normalize('NFKC', expression)
  return expression


# Examples
v1 = lexpr(r'x')
v2 = lexpr(r'x1')
v3 = lexpr(r'e1')
v4 = lexpr(r'z1')
v5 = lexpr(r'y')

atom1 = lexpr(r'_student(z1)')
atom2 = lexpr(r'_student(x,z1)')
atom3 = lexpr(r'(x = z1)')
atom4 = lexpr(r'_run(Subj(x))')
atom5 = lexpr(r'(Subj(x) = y)')

nonatom1 = lexpr(r'\x.\y.sees(x,y)(john)(mary)')
nonatom2 = lexpr(r'\x.sees(x,y)(john)')

and1 = lexpr(r'_student(z1) & _run(z1)')
and2 = lexpr(r'_student(z1) & _run(z1) & _boy(y)')
and3 = lexpr(r'_student(z1) & (x = z1)')

or1 = lexpr(r'_student(z1) | _run(z1)')
or2 = lexpr(r'(_student(z1) | _run(z1)) & _boy(y)')
or3 = lexpr(r'_student(z1) | (x = z1)')

imp1 = lexpr(r'_student(z1) -> _run(z1)')
imp2 = lexpr(r'(_student(z1) & _run(z1)) -> _boy(y)')
imp3 = lexpr(r'_student(z1) -> (x = z1)')

neg1 = lexpr(r'- _student(z1)')
neg2 = lexpr(r'- (_student(z1) & _run(z1))')
neg3 = lexpr(r'_student(z1) & - (x = z1)')

ex1 = lexpr(r'exists z1.(_student(z1) & _run(z1))')
ex2 = lexpr(r'exists z1.(_student(z1) & _run(z1) & exists x.(boy(x) & like(z1,x)))')
ex3 = lexpr(r'exists x z1.(_student(z1) & _run(z1) & boy(x) & like(z1,x))')
ex4 = lexpr(r'exists e x.(_run(z1) & boy(x) & like(z1,x))')
ex5 = lexpr(r'exists e x.(_run(e) & boy(x) & (Subj(e) = x))')
ex6 = lexpr(r'exists z1.(_student(z1) & exists x.(boy(x) & like(z1,x)))')
ex7 = lexpr(r'exists e.(_come(e) & (Subj(e) = x) & exists x.(_party(x) & _to(e,x)))')
ex8 = lexpr(r'exists e z1.(_come(z1) & (Subj(e) = z1) & exists z1.(_party(z1) & _to(e,z1)))')
ex9 = lexpr(r'exists e y.(_run(e) & boy(y) & (Subj(e) = y)) & exists e y.(_run(e) & boy(y) & (Subj(e) = y))')
ex10 = lexpr(r'exists e1 e2 x1 (_run(e1) & _walk(e2) & (Subj(e1) = x1) & (Subj(e2) = x1) & _boy(x1))')
ex11 = lexpr(r'exists e1 e2 x1 (_歩く(e1) & _走る(e2) & (Subj(e1) = x1) & (Subj(e2) = x1) & _男の子(x1))')

all1 = lexpr(r'forall x. (_student(x))')
all2 = lexpr(r'forall x. (_student(x) -> _run(x))')
all3 = lexpr(r'all x.(_student(x) -> exists e.(_come(e) & (Subj(e) = x) & exists x.(_party(x) & _to(e,x))))')

tr1 = lexpr(r'True')
tr2 = lexpr(r'True & (x = z1)')
tr3 = lexpr(r'_student(z1) & True & (x = z1)')

lam1 = lexpr(r'\x._student(x)')
lam2 = lexpr(r'know(john,\x._student(x))')
lam3 = lexpr(r'\x.\y.sees(x,y)(john)(mary)')
lam4 = lexpr(r'\x.\y.sees(x,y)(john, mary)')
lam5 = lexpr(r'all x.(man(x) & (\x.exists y.walks(x,y))(x))')
lam6 = lexpr(r'(\P.\Q.exists x.(P(x) & Q(x)))(\x.dog(x))(\x.bark(x))')

comp1 = lexpr(r'exists x.(_john(x) & True & exists z1.(_student(z1) & True & (x = z1)))')
comp2 = lexpr(r'exists x.(_john(x) & True & exists e.(_come(e) & (Subj(e) = x) & exists x.(_party(x) & True & _to(e,x) & True)))')
comp3 = lexpr(r'all x.(_student(x) -> (True -> exists e.(_come(e) & (Subj(e) = x) & exists x.(_party(x) & True & _to(e,x) & True))))')
comp4 = lexpr(r'exists x.(_john(x) & True & exists z1.(_student(z1) & True & (x = z1)) & exists x.(_walk(x) & _with(x,z1)))')
comp5 = lexpr(r'exists x.(_john(x) & forall z1.(_student(z1) & rel(x,z1)))')
comp6 = lexpr(r'(exists x.(_boy(x) & True & exists e.(_walk(e) & (Subj(e) = x) & True)) & exists x.(_girl(x) & True & exists e.(_walk(e) & (Subj(e) = x) & True)))')
comp7 = lexpr(r'(exists x.(_john(x) & True & exists z1.(_coffee(z1) & True & exists e.(_drink(e) & (Subj(e) = x) & (Acc(e) = z1) & True))) & exists x.(_bob(x) & True & exists z1.(_coffee(z1) & True & exists e.(_drink(e) & (Subj(e) = x) & (Acc(e) = z1) & True))))')
comp8 = lexpr(r'(exists x.(_boy(x) & True & exists z1.(_coffee(z1) & True & exists e.(_drink(e) & (Subj(e) = x) & (Acc(e) = z1) & True)) & exists z2.(_guitar(z2) & True & exists e.(_play(e) & (Subj(e) = x) & (Acc(e) = z2) & True))) & exists x.(_girl(x) & True & exists z1.(_coffee(z1) & True & exists e.(_drink(e) & (Subj(e) = x) & (Acc(e) = z1) & True)) & exists z2.(_guitar(z2) & True & exists e.(_play(e) & (Subj(e) = x) & (Acc(e) = z2) & True))))')
comp9 = lexpr(r'(exists x.(_boy(x) & True & exists e.(_walk(e) & (Subj(e) = x) & exists x.(_park(x) & True & _in(e,x) & True))) & exists x.(_girl(x) & True & exists e.(_walk(e) & (Subj(e) = x) & exists x.(_park(x) & True & _in(e,x) & True))))')
comp10 = lexpr(r'(exists x.(_john(x) & True & exists e.(_move(e) & (Subj(e) = x) & exists x.(_tokyo(x) & True & _from(e,x) & exists x.(_paris(x) & True & _to(e,x) & True)))) & exists x.(_bob(x) & True & exists e.(_move(e) & (Subj(e) = x) & exists x.(_tokyo(x) & True & _from(e,x) & exists x.(_paris(x) & True & _to(e,x) & True)))))')
comp11 = lexpr(r'(exists x.(_boy(x) & True & exists e.(_walk(e) & (Subj(e) = x) & _slowly(e) & True)) & exists x.(_girl(x) & True & exists e.(_walk(e) & (Subj(e) = x) & _slowly(e) & True)))')
comp12 = lexpr(r'(exists x.(_boy(x) & True & exists e.(_walk(e) & (Subj(e) = x) & _slowly(e) & exists x.(_park(x) & True & _in(e,x) & True))) & exists x.(_girl(x) & True & exists e.(_walk(e) & (Subj(e) = x) & _slowly(e) & exists x.(_park(x) & True & _in(e,x) & True))))')
comp13 = lexpr(r'exists x.(_walking(x) & _man(x) & True & exists e.((Subj(e) = Subj(e)) & True))')


test = [v1,v2,v3,
        atom1,atom2,atom3,atom4,atom5,
        nonatom1,nonatom2,
        and1,and2,and3,
        or1,or2,or3,
        imp1,imp2,imp3,
        neg1,neg2,neg3,
        ex1,ex2,ex3,ex4,ex5,ex6,ex7,ex8,ex9,
        all1,all2,all3,
        tr1,tr2,tr3,
        lam1,lam2,lam3,lam4,lam5,lam6,
        comp1,comp2,comp3,comp4,comp5,comp6,comp7,comp8,comp9,comp10,comp11,comp12,comp13]

def demo(function):
    for formula in test:
        answer = function(formula)
        print('Formula: {0}'.format(formula))
        print('Prediction: {0}\n'.format(answer))



# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

def normalize_token(token):
    """
    Convert symbols to avoid collisions with reserved punctuation
    in NLTK and coq.
    To avoid collisions with reserved words, we prefix each token
    with an underscore '_'.
    """
    normalized = token
    normalized = re.sub(r'\.', '_DOT', normalized)
    normalized = re.sub(r',', '_COMMA', normalized)
    normalized = re.sub(r'\(', '_LEFTB', normalized)
    normalized = re.sub(r'\)', '_RIGHTB', normalized)
    normalized = re.sub(r'^-$', '_HYPHEN', normalized)
    normalized = re.sub(r'^&$', '_AMPERSAND', normalized)
    normalized = re.sub(r'!', '_EXCLAMATION', normalized)
    normalized = re.sub(r'-', '_dash_', normalized)
    if not normalized.startswith('_'):
        normalized = '_' + normalized
    return normalized

def denormalize_token(token):
    """
    Unconvert symbols. This is the reverse operation as above.
    """
    denormalized = token
    denormalized = re.sub('_DOT', r'\.', denormalized)
    denormalized = re.sub('_COMMA', r',', denormalized)
    denormalized = re.sub('_LEFTB', r'\(', denormalized)
    denormalized = re.sub('_RIGHTB', r'\)', denormalized)
    denormalized = re.sub('_HYPHEN', r'^-$', denormalized)
    denormalized = re.sub('_AMPERSAND', r'^&$', denormalized)
    denormalized = re.sub('_EXCLAMATION', r'!', denormalized)
    denormalized = re.sub('_dash_', r'-', denormalized)
    # Remove possible suffix that was introduced to avoid type clashes.
    denormalized = re.sub(r'_[a-z][0-9]$', '', denormalized)
    denormalized = denormalized.lstrip('_')
    return denormalized





from nltk.sem.logic import *
from .logic_parser import lexpr
from .nltk2normal import remove_true


def run(expression):
    if not isinstance(expression, Expression):
        expression = lexpr(expression)
    expression = remove_true(expression)
    return _run(expression)


def _run(expression):
    if isinstance(expression, ApplicationExpression):
        return {
            'type': 'app',
            'func': _run(expression.function),
            'arg': _run(expression.argument)
            }
    elif isinstance(expression, EqualityExpression):
        return {
            'type': 'equal',
            'first': _run(expression.first),
            'second': _run(expression.second)
        }
    elif isinstance(expression, AndExpression):
        return {
            'type': 'and',
            'first': _run(expression.first),
            'second': _run(expression.second)
        }
    elif isinstance(expression, OrExpression):
        return {
            'type': 'or',
            'first': _run(expression.first),
            'second': _run(expression.second)
        }
    elif isinstance(expression, ImpExpression):
        return {
            'type': 'imp',
            'first': _run(expression.first),
            'second': _run(expression.second)
        }
    elif isinstance(expression, NegatedExpression):
        return {
            'type': 'not',
            'term': _run(expression.term),
        }
    elif isinstance(expression, ExistsExpression):
        return {
            'type': 'exists',
            'variable': str(expression.variable),
            'term': _run(expression.term),
        }
    elif isinstance(expression, AllExpression):
        return {
            'type': 'all',
            'variable': str(expression.variable),
            'term': _run(expression.term),
        }
    elif isinstance(expression, LambdaExpression):
        return {
            'type': 'lambda',
            'variable': str(expression.variable),
            'term': _run(expression.term),
        }
    else:
        return str(expression)



import os
import random
import chainer
import chainer.links as L
import chainer.functions as F
from chainer import cuda
from chainer import Variable
from chainer.dataset.convert import _concat_arrays
from depccg.utils import read_model_defs, normalize
from .biaffine import Biaffine, Bilinear
from depccg.chainer.param import Param
from .fixed_length_n_step_lstm import FixedLengthNStepLSTM
from depccg.types import ScoringResult


UNK = "*UNKNOWN*"
OOR2 = "OOR2"
OOR3 = "OOR3"
OOR4 = "OOR4"
START = "*START*"
END = "*END*"
IGNORE = -1
MISS = -2


def get_suffix(word):
    return [word[-1],
            word[-2:] if len(word) > 1 else OOR2,
            word[-3:] if len(word) > 2 else OOR3,
            word[-4:] if len(word) > 3 else OOR4]


def get_prefix(word):
    return [word[0],
            word[:2] if len(word) > 1 else OOR2,
            word[:3] if len(word) > 2 else OOR3,
            word[:4] if len(word) > 3 else OOR4]


class FeatureExtractor(object):
    def __init__(self, model_path, length=False):
        self.words = read_model_defs(os.path.join(model_path, "words.txt"))
        self.suffixes = read_model_defs(
            os.path.join(model_path, "suffixes.txt"))
        self.prefixes = read_model_defs(
            os.path.join(model_path, "prefixes.txt"))
        self.unk_word = self.words[UNK]
        self.start_word = self.words[START]
        self.end_word = self.words[END]
        self.unk_suf = self.suffixes[UNK]
        self.unk_prf = self.prefixes[UNK]
        self.start_pre = [[self.prefixes[START]] + [IGNORE] * 3]
        self.start_suf = [[self.suffixes[START]] + [IGNORE] * 3]
        self.end_pre = [[self.prefixes[END]] + [IGNORE] * 3]
        self.end_suf = [[self.suffixes[END]] + [IGNORE] * 3]
        self.length = length

    def process(self, words, xp):
        """
        words: list of unicode tokens
        """
        words = list(map(normalize, words))
        w = xp.array([self.start_word] + [self.words.get(
            x.lower(), self.unk_word) for x in words] + [self.end_word], 'i')
        s = xp.asarray(self.start_suf + [[self.suffixes.get(
            f, self.unk_suf) for f in get_suffix(x)] for x in words] + self.end_suf, 'i')
        p = xp.asarray(self.start_pre + [[self.prefixes.get(
            f, self.unk_prf) for f in get_prefix(x)] for x in words] + self.end_pre, 'i')
        if not self.length:
            return w, s, p
        else:
            return w, s, p, w.shape[0]


class Linear(L.Linear):

    def __call__(self, x):
        shape = x.shape
        if len(shape) == 3:
            x = F.reshape(x, (-1, shape[2]))
        y = super(Linear, self).__call__(x)
        if len(shape) == 3:
            y = F.reshape(y, (shape[0], shape[1], -1))
        return y


def concat_examples(batch, device=None):
    if len(batch) == 0:
        raise ValueError('batch is empty')

    if device is None:
        def to_device(x):
            return x
    elif device < 0:
        to_device = cuda.to_cpu
    else:
        def to_device(x):
            return cuda.to_gpu(x, device, cuda.Stream.null)

    result = [to_device(_concat_arrays([s[0] for s in batch], -1)),  # ws
              to_device(_concat_arrays([s[1] for s in batch], -1)),  # ps
              to_device(_concat_arrays([s[2] for s in batch], -1)),  # ss
              [s[3] for s in batch]]                                # ls

    if len(batch[0]) == 7:
        result.append([to_device(s[4]) for s in batch])            # cat_ts
        result.append([to_device(s[5]) for s in batch])            # dep_ts
        result.append(to_device(_concat_arrays(
            [s[6] for s in batch], None)))  # weights

    return tuple(result)


class FastBiaffineLSTMParser(chainer.Chain):
    def __init__(self, model_path, word_dim=None, afix_dim=None, nlayers=2,
                 hidden_dim=128, dep_dim=100, dropout_ratio=0.5):
        Param.load(self, os.path.join(model_path, 'tagger_defs.txt'))
        self.extractor = FeatureExtractor(model_path, length=True)
        self.in_dim = self.word_dim + 8 * self.afix_dim
        self.dropout_ratio = dropout_ratio
        super(FastBiaffineLSTMParser, self).__init__(
            emb_word=L.EmbedID(self.n_words, self.word_dim,
                               ignore_label=IGNORE),
            emb_suf=L.EmbedID(self.n_suffixes, self.afix_dim,
                              ignore_label=IGNORE),
            emb_prf=L.EmbedID(self.n_prefixes, self.afix_dim,
                              ignore_label=IGNORE),
            lstm_f=FixedLengthNStepLSTM(
                self.nlayers, self.in_dim, self.hidden_dim, 0.32),
            lstm_b=FixedLengthNStepLSTM(
                self.nlayers, self.in_dim, self.hidden_dim, 0.32),
            arc_dep=Linear(2 * self.hidden_dim, self.dep_dim),
            arc_head=Linear(2 * self.hidden_dim, self.dep_dim),
            rel_dep=Linear(2 * self.hidden_dim, self.dep_dim),
            rel_head=Linear(2 * self.hidden_dim, self.dep_dim),
            biaffine_arc=Biaffine(self.dep_dim),
            biaffine_tag=Bilinear(self.dep_dim, self.dep_dim, len(self.targets)))

    def forward(self, ws, ss, ps, ls, dep_ts=None):
        batchsize, slen = ws.shape

        wss = self.emb_word(ws)
        sss = F.reshape(self.emb_suf(ss), (batchsize, slen, 4 * self.afix_dim))
        pss = F.reshape(self.emb_prf(ps), (batchsize, slen, 4 * self.afix_dim))
        ins = F.dropout(F.concat([wss, sss, pss], 2), self.dropout_ratio)
        xs_f = F.transpose(ins, (1, 0, 2))
        xs_b = xs_f[::-1]

        cx_f, hx_f, cx_b, hx_b = self._init_state(batchsize)
        _, _, hs_f = self.lstm_f(hx_f, cx_f, xs_f, train=False)
        _, _, hs_b = self.lstm_b(hx_b, cx_b, xs_b, train=False)

        # (batch, length, hidden_dim)
        hs = F.transpose(F.concat([hs_f, hs_b[::-1]], 2), (1, 0, 2))

        dep_ys = self.biaffine_arc(
            F.elu(F.dropout(self.arc_dep(hs), 0.32)),
            F.elu(F.dropout(self.arc_head(hs), 0.32)))

        if dep_ts is not None and random.random >= 0.5:
            heads = dep_ts
        else:
            heads = F.flatten(F.argmax(dep_ys, axis=2)) + \
                self.xp.repeat(self.xp.arange(0, batchsize * slen, slen), slen)

        hs = F.reshape(hs, (batchsize * slen, -1))
        heads = F.permutate(
            F.elu(F.dropout(
                self.rel_head(hs), 0.32)), heads)

        childs = F.elu(F.dropout(self.rel_dep(hs), 0.32))
        cat_ys = self.biaffine_tag(childs, heads)

        dep_ys = F.split_axis(
            dep_ys, batchsize, 0) if batchsize > 1 else [dep_ys]
        dep_ys = [F.reshape(v, v.shape[1:])[:l, :l]
                  for v, l in zip(dep_ys, ls)]

        cat_ys = F.split_axis(
            cat_ys, batchsize, 0) if batchsize > 1 else [cat_ys]
        cat_ys = [v[:l] for v, l in zip(cat_ys, ls)]
        return cat_ys, dep_ys

    def _predict(self, xs):
        xs = [self.extractor.process(x, self.xp) for x in xs]
        ws, ss, ps, ls = concat_examples(xs)
        with chainer.no_backprop_mode(), chainer.using_config('train', False):
            cat_ys, dep_ys = self.forward(ws, ss, ps, ls)
        cat_ys = [F.log_softmax(y[1:-1]).data for y in cat_ys]
        cat_ys = [chainer.cuda.to_cpu(y) for y in cat_ys]
        dep_ys = [F.log_softmax(y[1:-1, :-1]).data for y in dep_ys]
        dep_ys = [chainer.cuda.to_cpu(y) for y in dep_ys]
        return list(zip(cat_ys, dep_ys))

    def predict_doc(self, doc, batchsize=32, gpu=-1):
        res = []
        doc = sorted(enumerate(doc), key=lambda x: len(x[1]))
        if gpu >= 0:
            chainer.cuda.get_device_from_id(gpu).use()
        for i in range(0, len(doc), batchsize):
            ids, batch = zip(*doc[i:i + batchsize])
            pred = self._predict(batch)
            res.extend((j, y) for j, y in zip(ids, pred))
        res = [ScoringResult(*pred) for _, pred in sorted(res)]
        return res, self.cats

    def _init_state(self, batchsize):
        res = [Variable(self.xp.zeros(
            (self.nlayers, batchsize, self.hidden_dim), 'f')) for _ in range(4)]
        return res

    @property
    def cats(self):
        return list(zip(*sorted(self.targets.items(), key=lambda x: x[1])))[0]



import math
import numpy as np

from chainer import cuda
from chainer import link
from chainer import functions as F
from chainer import initializer
from chainer.initializers.normal import HeNormal
from chainer.initializers.constant import Constant
from chainer.initializers.constant import Identity


def _get_initializer(initializer, scale=1.0):
    if initializer is None:
        return HeNormal(scale / np.sqrt(2))
    if np.isscalar(initializer):
        return Constant(initializer * scale)
    if isinstance(initializer, np.ndarray):
        return Constant(initializer * scale)

    assert callable(initializer)
    if scale == 1.0:
        return initializer
    return _ScaledInitializer(initializer, scale)


class _ScaledInitializer(initializer.Initializer):

    def __init__(self, initializer, scale=1.0):
        self.initializer = initializer
        self.scale = scale
        dtype = getattr(initializer, 'dtype', None)
        super(Identity, self).__init__(dtype)

    def __call__(self, array):
        self.initializer(array)
        array *= self.scale


class Biaffine(link.Link):

    def __init__(self, in_size, wscale=1,
                 initialW=None, initial_bias=None):
        super(Biaffine, self).__init__()

        self._W_initializer = _get_initializer(
            initialW, math.sqrt(wscale))

        self._initialize_params(in_size)

    def _initialize_params(self, in_size):
        self.add_param('W', (in_size + 1, in_size),
                       initializer=self._W_initializer)

    def forward_one(self, x1, x2):
        xp = cuda.get_array_module(x1.data)
        return F.matmul(
            F.concat([x1, xp.ones((x1.shape[0], 1), 'f')]),  # (slen, hidden+1)
            F.matmul(self.W, x2, transb=True))  # (hidden+1, hidden) * (slen, hidden)^T

    def forward_batch(self, x1, x2):
        xp = cuda.get_array_module(x1.data)
        batch, slen, hidden = x2.shape
        return F.batch_matmul(
            # (batch, slen, hidden+1)
            F.concat([x1, xp.ones((batch, slen, 1), 'f')], 2),
            F.reshape(F.linear(F.reshape(x2, (batch * slen, -1)), self.W),
                      (batch, slen, -1)), transb=True)

    def __call__(self, x1, x2):
        dim = len(x1.shape)
        if dim == 3:
            return self.forward_batch(x1, x2)
        elif dim == 2:
            return self.forward_one(x1, x2)
        else:
            raise RuntimeError()


class Bilinear(link.Link):

    # chainer.links.Bilinear may have some problem with GPU
    # and results in nan with batches with big size

    def __init__(self, in_size1, in_size2, out_size, wscale=1,
                 initialW=None, initial_bias=None, bias=0):
        super(Bilinear, self).__init__()

        self._W_initializer = _get_initializer(
            initialW, math.sqrt(wscale))
        if initial_bias is None:
            initial_bias = bias
        self.bias_initializer = _get_initializer(initial_bias)

        # same parameters as chainer.links.Bilinear
        # so that both can use serialized parameters of the other
        self.add_param('W', (in_size1, in_size2, out_size),
                       initializer=self._W_initializer)
        self.add_param('V1', (in_size1, out_size),
                       initializer=self._W_initializer)
        self.add_param('V2', (in_size2, out_size),
                       initializer=self._W_initializer)
        self.add_param('b', out_size,
                       initializer=self.bias_initializer)
        self.in_size1 = in_size1
        self.in_size2 = in_size2
        self.out_size = out_size

    def __call__(self, e1, e2):
        ele2 = F.reshape(
            F.batch_matmul(e1[:, :, None], e2[:, None, :]), (-1, self.in_size1 * self.in_size2))

        res = F.matmul(ele2,
                       F.reshape(self.W, (self.in_size1 * self.in_size2, self.out_size))) + \
            F.matmul(e1, self.V1) + \
            F.matmul(e2, self.V2)

        res, bias = F.broadcast(res, self.b)
        return res + bias



from typing import List, Tuple, Dict, NamedTuple, Optional
import json
import logging
import numpy
from depccg.types import ScoringResult

from depccg.cat import Category


logger = logging.getLogger(__name__)


def is_json(file_path: str) -> bool:
    try:
        with open(file_path, 'r') as data_file:
            json.load(data_file)
            return True
    except json.JSONDecodeError:
        return False


def normalize(word: str) -> str:
    if word == "-LRB-":
        return "("
    elif word == "-RRB-":
        return ")"
    elif word == "-LCB-":
        return "{"
    elif word == "-RCB-":
        return "}"
    elif word == "-LSB-":
        return "["
    elif word == "-RSB-":
        return "]"
    else:
        return word


def denormalize(word: str) -> str:
    if word == "(":
        return "-LRB-"
    elif word == ")":
        return "-RRB-"
    elif word == "{":
        return "-LCB-"
    elif word == "}":
        return "-RCB-"
    elif word == "[":
        return "-LSB-"
    elif word == "]":
        return "-RSB-"
    word = word.replace(">", "-RAB-")
    word = word.replace("<", "-LAB-")
    return word


def read_pretrained_embeddings(filepath: str) -> numpy.ndarray:
    nvocab = 0
    io = open(filepath)
    dim = len(io.readline().split())
    io.seek(0)
    for _ in io:
        nvocab += 1
    io.seek(0)
    res = numpy.empty((nvocab, dim), dtype=numpy.float32)
    for i, line in enumerate(io):
        line = line.strip()
        if len(line) == 0:
            continue
        res[i] = line.split()
    io.close()
    return res


def read_model_defs(filepath: str) -> Dict[str, int]:
    return {
        line.strip().split(' ')[0]: i
        for i, line in enumerate(open(filepath, encoding='utf-8'))
    }


def remove_comment(line: str) -> str:
    comment = line.find('#')
    if comment != -1:
        line = line[:comment]
    return line.strip()


class SpanInfo(NamedTuple):
    cat: Category
    idx: int
    end_idx: Optional[int] = None


def read_partial_tree(string: str) -> Tuple[List[str], List[SpanInfo]]:
    stack = []
    spans = []
    words = []
    buf = list(reversed(string.split()))
    counter = 0
    while buf:
        item = buf.pop()
        if item.startswith('<'):
            cat = item[1:]
            cat = None if cat == 'X' else Category.parse(cat)
            stack.append(cat)
            stack.append(counter)
        elif item == '>':
            start = stack.pop()
            cat = stack.pop()
            spans.append(SpanInfo(cat, start, counter - start))
        else:
            items = item.split('|')
            if len(items) == 1:
                words.append(items[0])
            elif len(items) == 2:
                cat, word = items
                assert len(cat) > 0 and len(word) > 0, \
                    'failed to parse partially annotated sentence.'
                words.append(word)
                spans.append(SpanInfo(Category.parse(cat), counter))
            counter += 1
    assert len(stack) == 0, 'failed to parse partially annotated sentence.'
    return words, spans


def maybe_split_and_join(string):
    if isinstance(string, list):
        split = string
        join = ' '.join(string)
    else:
        assert isinstance(string, str)
        split = string.split(' ')
        join = string
    return split, join


def read_weights(filename, file_type='json'):
    assert file_type == 'json'
    categories = None
    scores = []
    for line in open(filename):
        json_dict = json.loads(line.strip())

        if categories is None:
            categories = [
                Category.parse(cat)
                for cat in json_dict['categories']
            ]

        dep_scores = numpy.array(json_dict['heads']) \
            .reshape(json_dict['heads_shape']) \
            .astype(numpy.float32)
        tag_scores = numpy.array(json_dict['head_tags']) \
            .reshape(json_dict['head_tags_shape']) \
            .astype(numpy.float32)

        scores.append(
            ScoringResult(
                tag_scores,
                dep_scores
            )
        )

    return scores, categories



from typing import List
from lxml import etree
from depccg.tree import Tree, ScoredTree


def _process_tree(tree: Tree) -> etree.Element:

    def rec(node, parent):

        if node.is_leaf:
            leaf_node = etree.SubElement(parent, 'lf')
            start, token = tokens.pop(0)
            leaf_node.set('start', str(start))
            leaf_node.set('span', '1')
            leaf_node.set('cat', str(node.cat))
            for k, v in token.items():
                leaf_node.set(k, v)
        else:
            rule_node = etree.SubElement(parent, 'rule')
            rule_node.set('type', node.op_string)
            rule_node.set('cat', str(node.cat))
            for child in node.children:
                rec(child, rule_node)

    tokens = list(enumerate(tree.tokens))
    result = etree.Element("ccg")
    rec(tree, result)

    return result


def xml_of(
    nbest_trees: List[List[ScoredTree]],
) -> etree.Element:
    """convert parsing results to a XML etree.Element in a format commonly used by C&C.

    Args:
        nbest_trees (List[List[ScoredTree]]): parsing results

    Returns:
        etree.Element: XML object
    """

    candc_node = etree.Element('candc')
    for sentence_index, trees in enumerate(nbest_trees, 1):
        for tree_index, (tree, _) in enumerate(trees, 1):
            out = _process_tree(tree)
            out.set('sentence', str(sentence_index))
            out.set('id', str(tree_index))
            candc_node.append(out)

    return candc_node



import itertools

import numpy
import six
import binascii
import time
import os

from chainer.links import NStepLSTM
import chainer.functions as F

from chainer import cuda
from chainer.functions.activation import lstm
from chainer.functions.array import concat
from chainer.functions.array import reshape
from chainer.functions.array import split_axis
from chainer.functions.array import stack
from chainer.functions.connection import linear
from chainer.functions.noise import dropout
from chainer.functions.connection.n_step_lstm import NStepLSTM as NStepLSTMFunction


# if cuda.cudnn_enabled:
#     cudnn = cuda.cudnn
#     libcudnn = cuda.cudnn
#     _cudnn_version = libcudnn.getVersion()


class DropoutRandomStates(object):

    def __init__(self, seed):
        self._states = None

        if seed is None:
            try:
                seed_str = binascii.hexlify(os.urandom(8))
                seed = numpy.uint64(int(seed_str, 16))
            except NotImplementedError:
                seed = numpy.uint64(time.clock() * 1000000)
        else:
            seed = numpy.uint64(seed)

        self._seed = seed

    def create_dropout_states(self, dropout):
        handle = cudnn.get_handle()
        if self._states is None:
            self._states = cudnn.DropoutStates(handle, self._seed)
        # TODO(unno): Make a method to set dropout instead of calling API
        cudnn.set_dropout_descriptor(self._states._desc, handle, dropout)

        return self._states


_random_states = {}


def get_random_state():
    global _random_states
    dev = cuda.Device()
    rs = _random_states.get(dev.id, None)
    if rs is None:
        rs = DropoutRandomStates(os.getenv('CHAINER_SEED'))
        _random_states[dev.id] = rs
    return rs


class DropoutRandomStates(object):

    def __init__(self, seed):
        self._states = None

        if seed is None:
            try:
                seed_str = binascii.hexlify(os.urandom(8))
                seed = numpy.uint64(int(seed_str, 16))
            except NotImplementedError:
                seed = numpy.uint64(time.clock() * 1000000)
        else:
            seed = numpy.uint64(seed)

        self._seed = seed

    def create_dropout_states(self, dropout):
        handle = cudnn.get_handle()
        if self._states is None:
            self._states = cudnn.DropoutStates(handle, self._seed)
        # TODO(unno): Make a method to set dropout instead of calling API
        cudnn.set_dropout_descriptor(self._states._desc, handle, dropout)

        return self._states


_random_states = {}


def get_random_state():
    global _random_states
    dev = cuda.Device()
    rs = _random_states.get(dev.id, None)
    if rs is None:
        rs = DropoutRandomStates(os.getenv('CHAINER_SEED'))
        _random_states[dev.id] = rs
    return rs


class PointerArray(object):

    def __init__(self, lst, back_pointer):
        self._value = numpy.array(lst, dtype=numpy.intp)
        # Store back_pointer to prevent the GC removes the original variable
        self._back_pointer = back_pointer

    @property
    def data(self):
        return self._value.ctypes.data


def _split(inputs, pos):
    return inputs[:pos], inputs[pos:]


def _stack_weight(ws):
    # TODO(unno): Input of the current LSTM implementaiton is shuffled
    w = stack.stack(ws, axis=1)
    shape = w.shape
    return reshape.reshape(w, (shape[0] * shape[1],) + shape[2:])


class FixedLengthNStepLSTM(NStepLSTM):

    def __call__(self, hx, cx, xs, train=True):
        # xs: (length, batch, dim)

        ws = [[w.w0, w.w1, w.w2, w.w3, w.w4, w.w5, w.w6, w.w7] for w in self]
        bs = [[w.b0, w.b1, w.b2, w.b3, w.b4, w.b5, w.b6, w.b7] for w in self]

        hy, cy, ys = fixed_length_n_step_lstm(
            self.n_layers, self.dropout, hx, cx, ws, bs, xs,
            train=train)

        return hy, cy, ys


def _make_tensor_descriptor_array(xs, length):
    """Make an array of pointers denoting pointers of tensor descriptors.

    """
    descs = []
    batch_size = xs.shape[0] // length
    for i in range(length):
        x = xs[i*batch_size:(i+1)*batch_size]
        if x.ndim < 3:
            shape = x.shape + (1,) * (3 - x.ndim)
            x = x.reshape(shape)
        desc = cudnn.create_tensor_nd_descriptor(x)
        descs.append(desc)
    return PointerArray([d.value for d in descs], descs)


class FixedLengthNStepLSTMFunction(NStepLSTMFunction):

    def check_type_forward(self, in_types):
        pass

    def forward(self, inputs):
        (hx, cx), inputs = _split(inputs, 2)
        ws, inputs = _split(inputs, self.n_layers * 8)
        bs, inputs = _split(inputs, self.n_layers * 8)
        xs = inputs[0]

        hx = cuda.cupy.ascontiguousarray(hx)
        cx = cuda.cupy.ascontiguousarray(cx)
        xs = cuda.cupy.ascontiguousarray(xs)

        #x_desc = cudnn.create_tensor_nd_descriptor(x_list[0][..., None])
        x_desc = cudnn.create_tensor_nd_descriptor(cuda.cupy.copy(xs[0][..., None]))
        # NOTE: copy is necessary here, since xs in the original NStepLSTM implementation is a different array from xs_list[0]

        length = xs.shape[0]
        n_units = hx.shape[2]

        #xs = cuda.cupy.concatenate(x_list, axis=0)
        xs = xs.reshape(-1, xs.shape[2])    # (length * batch_size, dim)
        #ys = cuda.cupy.empty((len(xs), n_units), dtype=xs.dtype)
        ys = cuda.cupy.empty((xs.shape[0], n_units), dtype=xs.dtype)

        handle = cudnn.get_handle()
        self.handle = handle

        rnn_desc = cudnn.create_rnn_descriptor(
            n_units, self.n_layers, self.states.desc,
            libcudnn.CUDNN_LINEAR_INPUT, libcudnn.CUDNN_UNIDIRECTIONAL,
            libcudnn.CUDNN_LSTM, libcudnn.CUDNN_DATA_FLOAT)
        self.rnn_desc = rnn_desc

        #c_x_descs = _make_tensor_descriptor_array(x_list)
        c_x_descs = _make_tensor_descriptor_array(xs, length)
        hx_desc = cudnn.create_tensor_nd_descriptor(hx)
        cx_desc = cudnn.create_tensor_nd_descriptor(cx)

        weights_size = libcudnn.getRNNParamsSize(
            handle, rnn_desc.value, x_desc.value, libcudnn.CUDNN_DATA_FLOAT)
        w = cuda.cupy.empty((weights_size // 4, 1, 1), dtype=numpy.float32)
        w_desc = cudnn.create_filter_descriptor(w)

        for layer in six.moves.range(self.n_layers):
            for lin_layer_id in six.moves.range(8):
                mat = cudnn.get_rnn_lin_layer_matrix_params(
                    handle, rnn_desc, layer, x_desc, w_desc, w,
                    lin_layer_id)
                m = mat.reshape(mat.size)
                m[...] = ws[layer * 8 + lin_layer_id].ravel()
                bias = cudnn.get_rnn_lin_layer_bias_params(
                    handle, rnn_desc, layer, x_desc, w_desc, w,
                    lin_layer_id)
                b = bias.reshape(bias.size)
                b[...] = bs[layer * 8 + lin_layer_id]
        self.w = w
        self.w_desc = w_desc

        #sections = numpy.cumsum([len(x) for x in x_list[:-1]])
        #y_list = cuda.cupy.split(ys, sections)

        #c_y_descs = _make_tensor_descriptor_array(y_list)
        c_y_descs = _make_tensor_descriptor_array(ys, length)
        hy = cuda.cupy.empty_like(hx)
        cy = cuda.cupy.empty_like(cx)
        hy_desc = cudnn.create_tensor_nd_descriptor(hy)
        cy_desc = cudnn.create_tensor_nd_descriptor(cy)

        work_size = libcudnn.getRNNWorkspaceSize(
            handle, rnn_desc.value, length, c_x_descs.data)
        workspace = cuda.cupy.empty((work_size,), dtype='b')
        self.workspace = workspace

        if not self.train:
            libcudnn.RNNForwardInference(
                handle, rnn_desc.value, length,
                c_x_descs.data, xs.data.ptr, hx_desc.value, hx.data.ptr,
                cx_desc.value, cx.data.ptr, w_desc.value, w.data.ptr,
                c_y_descs.data, ys.data.ptr, hy_desc.value, hy.data.ptr,
                cy_desc.value, cy.data.ptr, workspace.data.ptr, work_size)

        else:
            reserve_size = libcudnn.getRNNTrainingReserveSize(
                handle, rnn_desc.value, length, c_x_descs.data)
            self.reserve_space = cuda.cupy.empty((reserve_size,), dtype='b')
            libcudnn.RNNForwardTraining(
                handle, rnn_desc.value, length,
                c_x_descs.data, xs.data.ptr, hx_desc.value, hx.data.ptr,
                cx_desc.value, cx.data.ptr, w_desc.value, w.data.ptr,
                c_y_descs.data, ys.data.ptr, hy_desc.value, hy.data.ptr,
                cy_desc.value, cy.data.ptr,
                workspace.data.ptr, work_size,
                self.reserve_space.data.ptr, reserve_size)

        self.c_y_descs = c_y_descs
        self.ys = ys
        self.c_x_descs = c_x_descs

        #return tuple([hy, cy] + y_list)
        return hy, cy, ys   # NOTE: ys has shape (-1, n_units)

    def backward(self, inputs, grads):
        (hx, cx), inputs = _split(inputs, 2)
        ws, inputs = _split(inputs, self.n_layers * 8)
        bs, inputs = _split(inputs, self.n_layers * 8)
        #x_list = inputs
        xs = inputs[0]

        hx = cuda.cupy.ascontiguousarray(hx)
        cx = cuda.cupy.ascontiguousarray(cx)
        xs = cuda.cupy.ascontiguousarray(xs)

        #dhy, dcy = grads[:2]
        #dy_list = list(grads[2:])
        dhy, dcy, dys = grads
        dys = cuda.cupy.ascontiguousarray(dys)
        if dhy is None:
            dhy = cuda.cupy.zeros_like(hx)
        if dcy is None:
            dcy = cuda.cupy.zeros_like(cx)
        #for i in six.moves.range(len(dys)):
        #    if dy_list[i] is None:
        #        dy_list[i] = cuda.cupy.zeros_like(x_list[i])

        #xs = cuda.cupy.concatenate(x_list, axis=0)
        length = xs.shape[0]
        batch_size = xs.shape[1]
        xs = xs.reshape(-1, xs.shape[2])
        #length = len(x_list)

        dhx = cuda.cupy.empty_like(hx)
        dcx = cuda.cupy.empty_like(cx)

        hx_desc = cudnn.create_tensor_nd_descriptor(hx)
        cx_desc = cudnn.create_tensor_nd_descriptor(cx)
        dhy_desc = cudnn.create_tensor_nd_descriptor(dhy)
        dcy_desc = cudnn.create_tensor_nd_descriptor(dcy)

        #c_dy_descs = _make_tensor_descriptor_array(dy_list)
        c_dy_descs = _make_tensor_descriptor_array(dys, length)
        #dys = cuda.cupy.concatenate(dy_list, axis=0)

        rnn_desc = self.rnn_desc
        handle = self.handle
        work_size = libcudnn.getRNNWorkspaceSize(
            handle, rnn_desc.value, length, self.c_x_descs.data)
        workspace = cuda.cupy.empty((work_size,), dtype='b')

        dhx_desc = cudnn.create_tensor_nd_descriptor(dhx)
        dcx_desc = cudnn.create_tensor_nd_descriptor(dcx)

        dxs = cuda.cupy.empty_like(xs)
        #sections = numpy.cumsum([len(x) for x in x_list[:-1]])
        #dx_list = cuda.cupy.split(dxs, sections, 0)
        #c_dx_descs = _make_tensor_descriptor_array(dx_list)
        c_dx_descs = _make_tensor_descriptor_array(dxs, length)

        libcudnn.RNNBackwardData(
            handle, rnn_desc.value, length,
            self.c_y_descs.data, self.ys.data.ptr,
            c_dy_descs.data, dys.data.ptr, dhy_desc.value, dhy.data.ptr,
            dcy_desc.value, dcy.data.ptr, self.w_desc.value, self.w.data.ptr,
            hx_desc.value, hx.data.ptr, cx_desc.value, cx.data.ptr,
            c_dx_descs.data, dxs.data.ptr, dhx_desc.value, dhx.data.ptr,
            dcx_desc.value, dcx.data.ptr, workspace.data.ptr, work_size,
            self.reserve_space.data.ptr, self.reserve_space.size)

        dw = cuda.cupy.zeros_like(self.w)
        dw_desc = cudnn.create_tensor_nd_descriptor(dw)
        libcudnn.RNNBackwardWeights(
            handle, rnn_desc.value, length,
            self.c_x_descs.data, xs.data.ptr,
            hx_desc.value, hx.data.ptr, self.c_y_descs.data, self.ys.data.ptr,
            workspace.data.ptr, work_size, dw_desc.value, dw.data.ptr,
            self.reserve_space.data.ptr, self.reserve_space.size)

        #dx = dx_list[0]
        dx = dxs[:batch_size]
        dx = dx.reshape(dx.shape + (1,))
        dx_desc = cudnn.create_tensor_nd_descriptor(dx)
        dws = [cuda.cupy.empty_like(w) for w in ws]
        dbs = [cuda.cupy.empty_like(b) for b in bs]
        for layer in six.moves.range(self.n_layers):
            for lin_layer_id in six.moves.range(8):
                mat = cudnn.get_rnn_lin_layer_matrix_params(
                    handle, rnn_desc, layer, dx_desc, dw_desc, dw,
                    lin_layer_id)
                v = dws[layer * 8 + lin_layer_id]
                v = v.reshape(v.size)
                v[:] = mat.ravel()
                bias = cudnn.get_rnn_lin_layer_bias_params(
                    handle, rnn_desc, layer, dx_desc, dw_desc, dw,
                    lin_layer_id)
                v = dbs[layer * 8 + lin_layer_id]
                v = v.reshape(v.size)
                v[:] = bias.ravel()

        dxs = dxs.reshape((length, batch_size, xs.shape[1]))

        #return tuple([dhx, dcx] + dws + dbs + dx_list)
        return tuple([dhx, dcx] + dws + dbs + [dxs])


def fixed_length_n_step_lstm(
        n_layers, dropout_ratio, hx, cx, ws, bs, xs, train=True,
        ):

    xp = cuda.get_array_module(hx, hx.data)

    if False:
    # if xp is not numpy and cuda.cudnn_enabled and _cudnn_version >= 5000:
        states = get_random_state().create_dropout_states(dropout_ratio)
        # flatten all input variables
        inputs = tuple(itertools.chain(
            (hx, cx),
            itertools.chain.from_iterable(ws),
            itertools.chain.from_iterable(bs),
            (xs,)))
        rnn = FixedLengthNStepLSTMFunction(n_layers, states, train=train)
        ret = rnn(*inputs)
        hy, cy, ys = ret
        _, batch_size, dim = hy.shape
        ys_reshape = F.reshape(ys, (-1, batch_size, dim))   # (length, batch, dim)
        return hy, cy, ys_reshape

    else:
        hx = split_axis.split_axis(hx, n_layers, axis=0, force_tuple=True)
        hx = [reshape.reshape(h, h.shape[1:]) for h in hx]
        cx = split_axis.split_axis(cx, n_layers, axis=0, force_tuple=True)
        cx = [reshape.reshape(c, c.shape[1:]) for c in cx]

        xws = [_stack_weight([w[2], w[0], w[1], w[3]]) for w in ws]
        hws = [_stack_weight([w[6], w[4], w[5], w[7]]) for w in ws]
        xbs = [_stack_weight([b[2], b[0], b[1], b[3]]) for b in bs]
        hbs = [_stack_weight([b[6], b[4], b[5], b[7]]) for b in bs]

        ys = []
        for x in xs:
            batch = x.shape[0]
            h_next = []
            c_next = []
            for layer in six.moves.range(n_layers):
                h = hx[layer]
                c = cx[layer]
                if h.shape[0] > batch:
                    h, h_rest = split_axis.split_axis(h, [batch], axis=0)
                    c, c_rest = split_axis.split_axis(c, [batch], axis=0)
                else:
                    h_rest = None

                x = dropout.dropout(x, ratio=dropout_ratio)
                h = dropout.dropout(h, ratio=dropout_ratio)
                lstm_in = linear.linear(x, xws[layer], xbs[layer]) + \
                          linear.linear(h, hws[layer], hbs[layer])

                c_bar, h_bar = lstm.lstm(c, lstm_in)
                if h_rest is not None:
                    h = concat.concat([h_bar, h_rest], axis=0)
                    c = concat.concat([c_bar, c_rest], axis=0)
                else:
                    h = h_bar
                    c = c_bar
                h_next.append(h)
                c_next.append(c)
                x = h_bar
            hx = h_next
            cx = c_next
            ys.append(x)

        hy = stack.stack(hx)
        cy = stack.stack(cx)
        #return hy, cy, tuple(ys)
        ys_concat = F.concat(ys, axis=0)
        ys_reshape = F.reshape(ys_concat, (-1, ys[0].shape[0], ys[0].shape[1]))     # (length, batch, dim)

        return hy, cy, ys_reshape



import torch

from allennlp.common import Params
from allennlp.data.vocabulary import Vocabulary
from allennlp.modules.token_embedders.embedding import Embedding
from allennlp.modules.time_distributed import TimeDistributed
from allennlp.modules.token_embedders.token_embedder import TokenEmbedder


@TokenEmbedder.register("afix_embedding")
class AfixEmbedding(TokenEmbedder):
    def __init__(self, embedding: Embedding, dropout: float = 0.0) -> None:
        super(AfixEmbedding, self).__init__()
        self._embedding = embedding
        if dropout > 0:
            self._dropout = torch.nn.Dropout(p=dropout)
        else:
            self._dropout = lambda x: x

    def get_output_dim(self) -> int:
        return self._embedding.get_output_dim() * 4

    def forward(self, token_afixes: torch.Tensor) -> torch.Tensor:
        batchsize, sentence_length, _ = token_afixes.size()
        embedding = TimeDistributed(self._embedding)
        embedding = TimeDistributed(embedding)
        embedded = embedding(token_afixes.unsqueeze(-1))
        return self._dropout(embedded.view(batchsize, sentence_length, -1))

    # The setdefault requires a custom from_params
    @classmethod
    def from_params(cls, vocab: Vocabulary, params: Params) -> 'AfixEmbedding':  # type: ignore
        # pylint: disable=arguments-differ
        embedding_params: Params = params.pop("embedding")
        # Embedding.from_params() uses "tokens" as the default namespace, but we need to change
        # that to be "token_characters" by default.
        embedding_params.setdefault("vocab_namespace", "afixes")
        embedding = Embedding.from_params(vocab, embedding_params)
        dropout = params.pop_float("dropout", 0.0)
        params.assert_empty(cls.__name__)
        return cls(embedding, dropout)



from typing import List, Tuple, Optional
import os
import subprocess
import tempfile
import logging
from pathlib import Path
from lxml import etree

from depccg.instance_models import MODEL_DIRECTORY
from depccg.morpha import MorphaStemmer
from depccg.types import Token

logger = logging.getLogger(__name__)


candc_cmd = "cat \"{0}\" | {1}/bin/pos --model {2} | {1}/bin/ner --model {3}"


def annotate_XX(sentences: List[List[str]], tokenize: bool = False) -> List[List[Token]]:
    if tokenize:
        raise NotImplementedError('no tokenizer implemented')

    return [
        [Token.of_word(word) for word in sentence]
        for sentence in sentences
    ]


def try_annotate_using_candc(sentences: List[List[str]], tokenize: bool = False) -> List[List[Token]]:
    if tokenize:
        raise NotImplementedError(
            'no tokenizer implemented in the C&C pipeline')

    candc_dir = os.environ.get('CANDC', None)
    candc_model_pos = None
    candc_model_ner = None
    fail = False
    if candc_dir:
        candc_dir = Path(candc_dir)
        candc_model_pos = Path(
            os.environ.get(
                'CANDC_MODEL_POS', str(candc_dir / 'models' / 'pos'))
        )
        candc_model_ner = Path(
            os.environ.get(
                'CANDC_MODEL_NER', str(candc_dir / 'models' / 'ner'))
        )

        if (
            (candc_dir / 'bin' / 'pos').exists()
            and (candc_dir / 'bin' / 'ner').exists()
            and candc_model_pos.exists()
            and candc_model_ner.exists()
        ):
            pass

        else:
            logger.info(
                'CANDC environmental variable may not be configured correctly.')
            logger.info(
                '$CANDC/bin/{pos,ner} and $CANDC/models/{pos,ner} are expected to exist.')
            fail = True

    else:
        fail = True

    if fail:
        logger.info('did not find C&C parser at CANDC environmental variable.')
        logger.info('fill POS tag etc. using XX tag.')
        return annotate_XX(sentences)

    logger.info('find C&C parser at CANDC environmental variable.')
    logger.info('use C&C pipeline to annotate POS and NER infos.')
    logger.info(f'C&C models: [{candc_model_pos}, {candc_model_ner}]')

    stemmer = MorphaStemmer(str(MODEL_DIRECTORY / 'verbstem.list'))

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'w') as f:
        for sentence in sentences:
            print(' '.join(sentence), file=f)

    command = candc_cmd.format(
        tmpfile,
        candc_dir,
        candc_model_pos,
        candc_model_ner
    )

    proc = subprocess.Popen(
        command,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    res, error = proc.communicate()
    try:
        tagged_sentences = res.decode('utf-8').strip().split('\n')
        tagged_sentences = [
            [
                tuple(token.split('|'))
                for token in sentence.strip().split(' ')
            ]
            for sentence in tagged_sentences
        ]
    except RuntimeError:
        raise RuntimeError(
            'failed to process C&C output. there might have been some problem '
            'during running C&C pipeline?\n'
            f'stderr:\n {error}'
        )

    results = []
    for sentence in tagged_sentences:
        words, poss = zip(*[(word, pos) for word, pos, _ in sentence])
        lemmas = stemmer.analyze(list(words), list(poss))
        tokens = [
            Token(
                word=word,
                pos=pos,
                entity=ner,
                lemma=lemma.lower(),
                chunk='XX'
            )
            for (word, pos, ner), lemma in zip(sentence, lemmas)
        ]
        results.append(tokens)
    return results


def annotate_using_spacy(
    sentences: List[List[str]],
    tokenize: bool = False,
    model_name: str = 'en_core_web_sm',
    n_threads: int = 2,
    batch_size: int = 10000
) -> Tuple[List[List[Token]], Optional[List[List[str]]]]:

    try:
        import spacy
        from spacy.tokens import Doc
    except ImportError:
        logger.error(
            'failed to import spacy. please install it by "pip install spacy".'
        )
        exit(1)

    nlp = spacy.load(model_name, disable=['parser'])
    logger.info('use spacy to annotate POS and NER infos.')

    if tokenize:
        docs = [nlp.tokenizer(' '.join(sentence)) for sentence in sentences]
    else:
        docs = [Doc(nlp.vocab, sentence) for sentence in sentences]

    for _, proc in nlp.pipeline:
        docs = proc.pipe(
            docs,
            batch_size=batch_size
        )

    results = []
    for sentence in docs:
        tokens = []
        for token in sentence:
            if token.ent_iob_ == 'O':
                ner = token.ent_iob_
            else:
                ner = token.ent_iob_ + '-' + token.ent_type_

            # takes care of pronoun
            if token.lemma_ == '-PRON-':
                lemma = str(token).lower()
            else:
                lemma = token.lemma_.lower()
            tokens.append(
                Token(
                    word=str(token),
                    pos=token.tag_,
                    entity=ner,
                    lemma=lemma,
                    chunk='XX'
                )
            )
        results.append(tokens)

    return results


def annotate_using_janome(
    sentences: List[List[str]],
    tokenize: bool = False
) -> Tuple[List[List[Token]], List[List[str]]]:

    assert tokenize, 'no support for using janome with pre-tokenized inputs'
    try:
        from janome.tokenizer import Tokenizer
    except ImportError:
        logger.error(
            'failed to import janome. please install it by "pip install janome".')
        exit(1)

    logger.info('use Janome to tokenize and annotate POS infos.')
    tokenizer = Tokenizer()
    results = []
    for sentence in sentences:
        sentence = ''.join(sentence)
        tokenized = list(tokenizer.tokenize(sentence))
        tokens = []
        for token in tokenized:
            pos, pos1, pos2, pos3 = token.part_of_speech.split(',')
            token = Token(
                word=token.surface,
                surf=token.surface,
                pos=pos,
                pos1=pos1,
                pos2=pos2,
                pos3=pos3,
                inflectionForm=token.infl_form,
                inflectionType=token.infl_type,
                reading=token.reading,
                base=token.base_form
            )
            tokens.append(token)
        results.append(tokens)
    return results


jigg_cmd = "java -Xmx2g -cp \"{0}/jar/*\" jigg.pipeline.Pipeline -annotators {1} -file {2} -output {3}"


def annotate_using_jigg(
    sentences: List[List[str]],
    tokenize: bool = False,
    pipeline: str = 'ssplit,kuromoji'
) -> Tuple[List[List[Token]], List[List[str]]]:

    assert tokenize, 'no support for using jigg with pre-tokenized inputs'
    logger.info('use Jigg to tokenize and annotate POS infos.')

    jigg_dir = os.environ.get('JIGG', None)
    if not jigg_dir:
        logger.error(
            'did not find Jigg at JIGG environmental variable. exiting..')
        exit(1)

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'w') as f:
        for sentence in sentences:
            print(' '.join(sentence), file=f)

    outfile = tempfile.mktemp()

    command = jigg_cmd.format(
        jigg_dir,
        pipeline,
        tmpfile,
        outfile
    )

    proc = subprocess.Popen(
        command,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    proc.communicate()
    results = []
    for sentence in etree.parse(outfile).getroot().xpath('*//sentence'):
        tokens = []
        for token in sentence.xpath('*//token'):
            attrib = token.attrib
            token = Token(
                word=attrib['surf'],
                surf=attrib['surf'],
                pos=attrib['pos'],
                pos1=attrib['pos1'],
                pos2=attrib['pos2'],
                pos3=attrib['pos3'],
                inflectionForm=attrib['inflectionForm'],
                inflectionType=attrib['inflectionType'],
                reading=attrib['reading'],
                base=attrib['base']
            )
            tokens.append(token)
        results.append(tokens)
    return results


english_annotator = {
    'candc': try_annotate_using_candc,
    'spacy': annotate_using_spacy,
}


japanese_annotator = {
    'janome': annotate_using_janome,
    'jigg': annotate_using_jigg,
}



from functools import partial
from collections import defaultdict

from allennlp.common.params import Params
from depccg.lang import get_global_language
from depccg.instance_models import GRAMMARS
from depccg.cat import Category


def read_params(
    param_path: str,
    disable_category_dictionary: bool = False,
    disable_seen_rules: bool = False,
):
    lang = get_global_language()
    params = Params.from_file(param_path)

    unary_rules = defaultdict(list)
    for key, value in params.pop('unary_rules'):
        unary_rules[Category.parse(key)].append(Category.parse(value))

    if disable_category_dictionary:
        category_dict = None
    else:
        category_dict = {
            word: [Category.parse(cat) for cat in cats]
            for word, cats in params.pop('cat_dict').items()
        }

    if disable_seen_rules:
        seen_rules = None
    else:
        seen_rules = {
            (Category.parse(x).clear_features('X', 'nb'),
             Category.parse(y).clear_features('X', 'nb'))
            for x, y in params.pop('seen_rules')
        }
        if len(seen_rules) == 0:
            seen_rules = None
    try:
        apply_binary_rules = partial(
            GRAMMARS[lang].apply_binary_rules,
            seen_rules=seen_rules
        )
        apply_unary_rules = partial(
            GRAMMARS[lang].apply_unary_rules,
            unary_rules=unary_rules
        )
    except KeyError:
        raise KeyError('unsupported language: {args.lang}')

    root_categories = [
        Category.parse(category)
        for category in params.pop('targets')
    ]

    return (
        apply_binary_rules,
        apply_unary_rules,
        category_dict,
        root_categories
    )




from .reader import read_trees_guess_extension
from ..printer import to_mathml
import logging
import argparse
from pathlib import Path

logger = logging.getLogger(__name__)


index_html = '''\
<html>
<head>
  <title>parse results</title>
</head>
<body>
<table border=1>
  <tr><td>id</td><td>file</td><td>sentence</td></tr>
  {}
</table>
</body>
</html>
'''

def to_mathml_separate_files(trees, directory_name='results'):
    out_directory = Path(directory_name)
    if out_directory.exists():
        raise RuntimeError(f'Directory "{directory_name}" already exists. Use other name.')
    out_directory.mkdir()
    trs = []
    for i, nbest in enumerate(trees):
        filename = f'{i}.html'
        mathml_html = to_mathml([nbest])
        with (out_directory / filename).open('w') as f:
            print(mathml_html, file=f)
        words = nbest[0][0].word if isinstance(nbest[0], tuple) else nbest[0].word
        trs.append(f'<tr><td>{i}</td><td><a href="{filename}">{filename}</a></td><td>{words}</td></tr>')
    with (out_directory / 'index.html').open('w') as f:
        print(index_html.format('\n'.join(trs)), file=f)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('PATH', help='path to either of *.auto, *.xml, *.jigg.xml, *.ptb')
    parser.add_argument('--mkdir', default=None, help='if specified, creates a directory')
    args = parser.parse_args()

    trees = [[tree] for _, _, tree in read_trees_guess_extension(args.PATH)]
    if not args.mkdir:
        print(to_mathml(trees))
    else:
        to_mathml_separate_files(trees, args.mkdir)



from depccg.cat import Category
from depccg.unification import Unification
import pytest


def test_basic():
    uni = Unification("(((a/b)/c)/d)/e", "f")
    x = Category.parse("(((a/b)/c)/d)/e")
    y = Category.parse("f")
    assert uni(x, y)
    assert uni["a"] == "a"
    assert uni["b"] == "b"
    assert uni["c"] == "c"
    assert uni["d"] == "d"
    assert uni["e"] == "e"
    assert uni["f"] == "f"
    with pytest.raises(RuntimeError, match="cannot use the same *"):
        uni(x, y)


def test_deep():
    uni = Unification("a/b", "c")
    x = Category.parse("(((a/b)/c)/d)/e")
    y = Category.parse("f")
    assert uni(x, y)
    assert uni["a"] == "((a/b)/c)/d"
    assert uni["b"] == "e"
    assert uni["c"] == "f"


def test_english():
    uni = Unification("a/b", "b")
    x = Category.parse("S[X]/NP[X]")
    y = Category.parse("NP[mod]")
    assert uni(x, y)
    assert uni["a"] == Category.parse('S[mod]')


def test_japanese():
    uni = Unification("(a\\b)/c", "c")

    x = Category.parse(
        "(S[mod=nm,form=base,fin=f]\\S[mod=nm,form=base,fin=f])/S[mod=nm,form=base,fin=f]")
    y = Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni(x, y)
    assert uni["a"] == Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni["b"] == Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni["c"] == Category.parse("S[mod=nm,form=base,fin=f]")

    # three variables
    uni = Unification("(a\\b)/c", "c")
    x = Category.parse(
        "(S[mod=X1,form=X2,fin=X3]\\S[mod=X1,form=X2,fin=X3])/S[mod=X1,form=X2,fin=X3]")
    y = Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni(x, y)
    assert uni["a"] == Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni["b"] == Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni["c"] == Category.parse("S[mod=nm,form=base,fin=f]")

    # only two variables
    uni = Unification("(a\\b)/c", "c")
    x = Category.parse(
        "(S[mod=X1,form=X2,fin=f]\\S[mod=X1,form=X2,fin=f])/S[mod=X1,form=X2,fin=f]")
    y = Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni(x, y)
    assert uni["a"] == Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni["b"] == Category.parse("S[mod=nm,form=base,fin=f]")
    assert uni["c"] == Category.parse("S[mod=nm,form=base,fin=f]")



from typing import List
from lxml import etree

from depccg.tree import ScoredTree, Tree
from depccg.cat import Category, TernaryFeature, UnaryFeature


def _cat_multi_valued(cat: Category) -> str:
    def rec(x: Category):
        if x.is_atomic:
            if isinstance(x.feature, UnaryFeature):
                if x.feature.value is None:
                    return x.base
                else:
                    return f'{x.base}[{x.feature}=true]'
            elif isinstance(x.feature, TernaryFeature):
                return str(x)
            else:
                raise RuntimeError(
                    f'unsupported feature type: {type(x.feature)}')
        else:
            return f'({_cat_multi_valued(x)})'

    if cat.is_atomic:
        return rec(cat)
    return f'{rec(cat.left)}{cat.slash}{rec(cat.right)}'


class _ConvertToJiggXML(object):
    def __init__(self, sid: int, use_symbol: bool) -> None:
        self.sid = sid
        self._spid = -1
        self.processed = 0
        self.use_symbol = use_symbol

    @property
    def spid(self) -> int:
        self._spid += 1
        return self._spid

    def process(self, tree: Tree, score: float = None) -> None:
        counter = 0

        def traverse(node: Tree) -> None:
            nonlocal counter
            id = f's{self.sid}_sp{self.spid}'
            xml_node = etree.SubElement(res, 'span')
            xml_node.set('category', _cat_multi_valued(node.cat))
            xml_node.set('id', id)
            if node.is_leaf:
                start_of_span = counter
                counter += 1
                xml_node.set('terminal', f's{self.sid}_{start_of_span}')
            else:
                childid, start_of_span = traverse(node.left_child)
                if not node.is_unary:
                    tmp, _ = traverse(node.right_child)
                    childid += ' ' + tmp
                xml_node.set('child', childid)
                xml_node.set(
                    'rule', node.op_symbol if self.use_symbol else node.op_string
                )
            xml_node.set('begin', str(start_of_span))
            xml_node.set('end', str(start_of_span + len(node)))
            return id, start_of_span

        res = etree.Element('ccg')
        res.set('id', f's{self.sid}_ccg{self.processed}')
        id, _ = traverse(tree)
        res.set('root', str(id))
        res[0].set('root', 'true')
        if score is not None:
            res.set('score', str(score))
        self.processed += 1
        return res


def to_jigg_xml(
    trees: List[List[ScoredTree]],
    use_symbol: bool = False
) -> etree.Element:
    """generate etree.Element XML object in jigg format
    containing all the parse results

    Args:
        trees (List[List[ScoredTree]]): parsing result
        use_symbol (bool, optional): [description]. Defaults to False.

    Returns:
        etree.Element: jigg format etree.Element tree
    """

    root_node = etree.Element('root')
    document_node = etree.SubElement(root_node, 'document')
    sentences_node = etree.SubElement(document_node, 'sentences')

    for sentence_index, parsed in enumerate(trees):

        sentence_node = etree.SubElement(sentences_node, 'sentence')
        tokens_node = etree.SubElement(sentence_node, 'tokens')
        cats = [leaf.cat for leaf in parsed[0].tree.leaves]
        tokens = parsed[0].tree.tokens

        for token_index, (token, cat) in enumerate(zip(tokens, cats)):
            token_node = etree.SubElement(tokens_node, 'token')
            token_node.set('start', str(token_index))
            token_node.set('cat', str(cat))
            token_node.set('id', f's{sentence_index}_{token_index}')

            if 'word' in token:
                token['surf'] = token.pop('word')
            if 'lemma' in token:
                token['base'] = token.pop('lemma')
            for k, v in token.items():
                token_node.set(k, v)

        converter = _ConvertToJiggXML(sentence_index, use_symbol)
        for tree, score in parsed:
            sentence_node.append(converter.process(tree, score))

    return root_node



from libcpp.string cimport string
from libcpp cimport bool
from libc.stdlib cimport malloc, free


cdef extern from "c/morpha.h":
    void morph_initialise(const char* filename)
    int morph_analyse(char* buffer, const char *str, int tag)


cdef unsigned MAX_MORPHA_LEN = 32;


cdef class MorphaStemmer:
    def __cinit__(self, str filename):
        cdef string c_filename = filename.encode('utf-8')
        morph_initialise(c_filename.c_str())

    cdef bool use_morpha(self, str word, str pos):
        if len(word) >= MAX_MORPHA_LEN:
            return False

        if pos == "NNP" or pos == "NNPS":
            return False

        if any(not (chr.isalpha() or chr == '-') for chr in word):
            return False

        return True

    cdef string analyze_one(self, string &word, string &pos, char *out):
        cdef char under_bar = b'_'
        cdef string inp = word
        inp += under_bar
        inp += pos
        morph_analyse(out, inp.c_str(), True)

    def analyze(self, list words, list poss):
        assert len(words) == len(poss)
        assert all(isinstance(word, str) for word in words)
        assert all(isinstance(pos, str) for pos in poss)
        cdef str word, pos
        cdef bytes out
        cdef string c_word, c_pos
        cdef char *buffer = <char *> malloc(MAX_MORPHA_LEN * sizeof(char))
        res = []
        for word, pos in zip(words, poss):
            if self.use_morpha(word, pos):
                c_word = word.encode('utf-8')
                c_pos = pos.encode('utf-8')
                self.analyze_one(c_word, c_pos, buffer)
                out = <bytes> buffer
                res.append(out.decode('utf-8'))
            else:
                res.append(word)
        free(buffer)
        return res





import math
import torch
from torch.nn.parameter import Parameter
import torch.nn.functional as F
from torch.nn.modules import Module


class BilinearWithBias(Module):
    def __init__(self, in1_features, in2_features, out_features):
        super(BilinearWithBias, self).__init__()
        self.in1_features = in1_features
        self.in2_features = in2_features
        self.out_features = out_features
        self.W = Parameter(torch.Tensor(out_features, in1_features, in2_features))
        self.V1 = Parameter(torch.Tensor(out_features, in1_features))
        self.V2 = Parameter(torch.Tensor(out_features, in2_features))
        self.bias = Parameter(torch.Tensor(out_features))
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.W.size(1))
        self.W.data.uniform_(-stdv, stdv)
        self.V1.data.uniform_(-stdv, stdv)
        self.V2.data.uniform_(-stdv, stdv)
        self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input1, input2):
        result = F.bilinear(input1, input2, self.W, self.bias)
        result += F.linear(input1, self.V1, None)
        result += F.linear(input2, self.V2, None)
        return result

    def extra_repr(self):
        return 'in1_features={}, in2_features={}, out_features={}, bias={}'.format(
            self.in1_features, self.in2_features, self.out_features, self.bias is not None
        )




import argparse
import logging
from .reader import read_trees_guess_extension
from depccg.printer.html import (
    _mathml_subtree, _MATHML_SUBTREE_TERMINAL, _mathml_cat, _MATHML_SUBTREE_NONTERMINAL, _MATHML_MAIN
)


logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                    level=logging.INFO)
logger = logging.getLogger(__name__)


def diff(tree1, tree2):
    def rec(node1, node2):
        nodes_are_different = False
        if node1.cat != node2.cat:
            nodes_are_different = True
        elif len(node1.children) != len(node2.children):
            nodes_are_different = True
        elif any(child1.word != child2.word
                 for child1, child2 in zip(node1.children, node2.children)):
            nodes_are_different = True

        if nodes_are_different:
            return (nodes_are_different,
                    _mathml_subtree(node1, bgcolor='peachpuff'),
                    _mathml_subtree(node2, bgcolor='peachpuff'))
        elif node1.is_leaf:
            assert node2.is_leaf
            node1_str = _MATHML_SUBTREE_TERMINAL.format(
                node1.word, _mathml_cat(str(node1.cat)))
            node2_str = _MATHML_SUBTREE_TERMINAL.format(
                node2.word, _mathml_cat(str(node2.cat)))
            return nodes_are_different, node1_str, node2_str
        else:
            children_are_different, node1_children, node2_children = \
                zip(*[rec(child1, child2)
                    for child1, child2 in zip(node1.children, node2.children)])
            node1_children = ''.join(node1_children)
            node2_children = ''.join(node2_children)
            node1_str = _MATHML_SUBTREE_NONTERMINAL.format(
                node1_children, _mathml_cat(str(node1.cat)), node1.op_string, '')
            node2_str = _MATHML_SUBTREE_NONTERMINAL.format(
                node2_children, _mathml_cat(str(node2.cat)), node2.op_string, '')
            nodes_are_different = any(children_are_different)
            return nodes_are_different, node1_str, node2_str

    return rec(tree1, tree2)


def to_diffs(tree_pairs, max_output_num, sampling):
    diffs = [(i, name1, name2, tree1, tree2) + diff(tree1, tree2)
             for i, ((name1, tree1), (name2, tree2)) in enumerate(tree_pairs)]
    diffs = [(i, name1, name2, tree1, tree2, tree1_str, tree2_str)
             for (i, name1, name2, tree1, tree2, trees_are_different, tree1_str, tree2_str) in diffs
             if trees_are_different]
    logger.info(f'The number of input tree pairs: {len(tree_pairs)}')
    logger.info(f'The number of different tree pairs: {len(diffs)}')
    if len(diffs) > max_output_num:
        logger.info(
            f'The number of different tree pairs exceeds --max-output-num: {max_output_num}.')
        logger.info(f'sample the subset of pairs using "{sampling}" method.')
        if sampling == 'head':
            diffs = diffs[:max_output_num]
        elif sampling == 'tail':
            diffs = diffs[-max_output_num:]
        elif sampling == 'random':
            import random
            indices = list(range(len(diffs)))
            indices = sorted(random.sample(indices, max_output_num))
            diffs = [diffs[i] for i in indices]
        else:
            assert False
    result = ''
    for (i, name1, name2, tree1, tree2, tree1_str, tree2_str) in diffs:
        if i > 0:
            result += '<hr>'
        result += f'<p><strong>{name1}</strong> ID={i}: {tree1.word}</p>'
        result += f'<math xmlns="http://www.w3.org/1998/Math/MathML">{tree1_str}</math>'
        result += f'<p><strong>{name2}</strong> ID={i}: {tree2.word}</p>'
        result += f'<math xmlns="http://www.w3.org/1998/Math/MathML">{tree2_str}</math>'
    return _MATHML_MAIN.format(result)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        'show diffs between the structures of two CCG trees')
    parser.add_argument('file1')
    parser.add_argument('file2', nargs='?', default=None)
    parser.add_argument('--max-output-num', default=50, type=int)
    parser.add_argument('--sampling', default='head',
                        choices=['head', 'tail', 'random'])
    args = parser.parse_args()
    if args.file2:
        file1_trees = [(args.file1, tree)
                       for _, _, tree in read_trees_guess_extension(args.file1)]
        file2_trees = [(args.file2, tree)
                       for _, _, tree in read_trees_guess_extension(args.file2)]
        assert len(file1_trees) == len(file2_trees)
        tree_pairs = list(zip(file1_trees, file2_trees))
    else:
        file1_trees = ((args.file1, tree)
                       for _, tree in read_trees_guess_extension(args.file1))
        tree_pairs = list(zip(file1_trees, file1_trees))

    print(to_diffs(tree_pairs, args.max_output_num, args.sampling))



# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging

from nltk.sem.logic import LogicParser
from nltk.sem.logic import LogicalExpressionException

logic_parser = LogicParser(type_check=False)
def lexpr(formula_str):
    try:
        expr = logic_parser.parse(formula_str)
    except LogicalExpressionException as e:
        msg = 'Failed to parse {0}. Error: {1}'.format(formula_str, e)
        logging.error(msg)
        raise LogicalExpressionException(0, msg)
    return expr




import os
import numpy as np
import chainer
import chainer.links as L
import chainer.functions as F

from depccg.utils import read_model_defs
from depccg.chainer.biaffine import Biaffine, Bilinear
from depccg.chainer.param import Param
from depccg.types import ScoringResult


UNK = "*UNKNOWN*"
START = "*START*"
END = "*END*"
IGNORE = -1


class FeatureExtractor(object):
    def __init__(self, model_path):
        self.model_path = model_path
        self.words = read_model_defs(os.path.join(model_path, 'words.txt'))
        self.chars = read_model_defs(os.path.join(model_path, 'chars.txt'))
        self.unk_word = self.words[UNK]
        self.start_word = self.words[START]
        self.end_word = self.words[END]
        self.unk_char = self.chars[UNK]
        self.start_char = self.chars[START]
        self.end_char = self.chars[END]

    def process(self, words, xp=np):
        """
        words: list of unicode tokens
        """
        w = xp.array([self.start_word] + [self.words.get(
            x, self.unk_word) for x in words] + [self.end_word], 'i')
        length = max(len(x) for x in words)
        c = -xp.ones((len(words) + 2, length), 'i')
        c[0, 0] = self.start_char
        c[-1, 0] = self.end_char
        for i, word in enumerate(words, 1):
            for j in range(len(word)):
                c[i, j] = self.chars.get(word[j], self.unk_char)
        return w, c, xp.array([length], 'i')


class BiaffineJaLSTMParser(chainer.Chain):
    def __init__(self, model_path):
        Param.load(self, os.path.join(model_path, 'tagger_defs.txt'))
        self.extractor = FeatureExtractor(model_path)
        self.in_dim = self.word_dim + self.char_dim
        super(BiaffineJaLSTMParser, self).__init__(
            emb_word=L.EmbedID(self.n_words, self.word_dim),
            emb_char=L.EmbedID(self.n_chars, 50, ignore_label=IGNORE),
            conv_char=L.Convolution2D(
                1, self.char_dim, (3, 50), stride=1, pad=(1, 0)),
            lstm_f=L.NStepLSTM(self.nlayers, self.in_dim,
                               self.hidden_dim, 0.32),
            lstm_b=L.NStepLSTM(self.nlayers, self.in_dim,
                               self.hidden_dim, 0.32),
            arc_dep=L.Linear(2 * self.hidden_dim, self.dep_dim),
            arc_head=L.Linear(2 * self.hidden_dim, self.dep_dim),
            rel_dep=L.Linear(2 * self.hidden_dim, self.dep_dim),
            rel_head=L.Linear(2 * self.hidden_dim, self.dep_dim),
            biaffine_arc=Biaffine(self.dep_dim),
            biaffine_tag=Bilinear(self.dep_dim, self.dep_dim, len(self.targets)))

    def forward(self, ws, cs, ls, dep_ts=None):
        ws = map(self.emb_word, ws)
        cs = [F.squeeze(
            F.max_pooling_2d(
                self.conv_char(
                    F.expand_dims(
                        self.emb_char(c), 1)), (int(l[0]), 1)))
              for c, l in zip(cs, ls)]
        xs_f = [F.dropout(F.concat([w, c]), 0.5) for w, c in zip(ws, cs)]
        xs_b = [x[::-1] for x in xs_f]

        _, _, hs_f = self.lstm_f(None, None, xs_f)
        _, _, hs_b = self.lstm_b(None, None, xs_b)
        hs_b = [x[::-1] for x in hs_b]
        hs = [F.concat([h_f, h_b]) for h_f, h_b in zip(hs_f, hs_b)]

        dep_ys = [self.biaffine_arc(
            F.elu(F.dropout(self.arc_dep(h), 0.32)),
            F.elu(F.dropout(self.arc_head(h), 0.32))) for h in hs]

        if dep_ts is not None:
            heads = dep_ts
        else:
            heads = [F.argmax(y, axis=1) for y in dep_ys]

        cat_ys = [self.biaffine_tag(
            F.elu(F.dropout(self.rel_dep(h), 0.32)),
            F.elu(F.dropout(self.rel_head(
                F.embed_id(t, h, ignore_label=IGNORE)), 0.32)))
            for h, t in zip(hs, heads)]

        return cat_ys, dep_ys

    def _predict(self, xs):
        xs = [self.extractor.process(x, self.xp) for x in xs]
        ws, ss, ps = zip(*xs)
        with chainer.no_backprop_mode(), chainer.using_config('train', False):
            cat_ys, dep_ys = self.forward(ws, ss, ps)
        cat_ys = [F.log_softmax(y[1:-1]).data for y in cat_ys]
        cat_ys = [chainer.cuda.to_cpu(y) for y in cat_ys]
        dep_ys = [F.log_softmax(y[1:-1, :-1]).data for y in dep_ys]
        dep_ys = [chainer.cuda.to_cpu(y) for y in dep_ys]
        return list(ScoringResult(*result) for result in zip(cat_ys, dep_ys))

    def predict_doc(self, doc, batchsize=16, gpu=-1):
        res = []
        if gpu >= 0:
            chainer.cuda.get_device_from_id(gpu).use()
        for i in range(0, len(doc), batchsize):
            res.extend(self._predict(doc[i:i + batchsize]))
        return res, self.cats

    @property
    def cats(self):
        return list(zip(*sorted(self.targets.items(), key=lambda x: x[1])))[0]



from typing import List
from io import StringIO

from depccg.cat import Category
from depccg.tree import ScoredTree, Tree


def _prolog_category_string(cat: Category) -> str:

    def rec(this_cat: Category) -> str:

        if this_cat.is_atomic:
            base = this_cat.base.lower()
            if base == '.':
                return "period"
            elif base == ",":
                return "comma"
            elif base == ":":
                return "colon"
            elif base == ";":
                return "semicolon"
            elif str(this_cat.feature) == "":
                return base
            else:
                return f'{base}:{str(this_cat.feature)}'

        else:
            left = rec(this_cat.left)
            right = rec(this_cat.right)
            return f"({left}{this_cat.slash}{right})"

    return rec(cat)


def _escape_prolog(text: str) -> str:
    return text.replace("'", "\\'")


_op_mapping = {
    'fa': "fa(",
    'ba': "ba(",
    'fx': "fc(",
    'fc': "fc(",
    'bx': "bxc(",
    'gfc': "gfc(",
    'gbx': "gbx(",
    'rp': "rp(",
    'lp': "lx(",
    'conj': "conj(",
    'conj2': "conj(",
}


def _prolog_string(tree: Tree, sentence_index: int) -> str:

    position = 0
    depth = 1

    def indent(output):
        output.write(" " * depth)

    def rec(node: Tree, output):
        nonlocal depth, position

        indent(output)
        if node.is_leaf:
            token = node.token
            result_str = (
                f"t({_prolog_category_string(node.cat)}, "
                f"\'{_escape_prolog(node.word)}\', "
                f"\'{_escape_prolog(token.lemma)}\', "
                f"\'{token.pos}\', "
                f"\'{token.chunk}\', "
                f"\'{token.entity}\')"
            )
            output.write(result_str)
            position += 1
        elif node.is_unary:
            this_cat = _prolog_category_string(node.cat)
            child_cat = _prolog_category_string(node.left_child.cat)
            output.write(f"lx({this_cat}, {child_cat},\n")
            depth += 1
            rec(node.child, output)
            depth -= 1
            output.write(")")
        else:
            output.write(_op_mapping[node.op_string])
            output.write(_prolog_category_string(node.cat))
            output.write(",")

            if node.op_string == 'conj2':
                cat_str = _prolog_category_string(node.right_child.cat)
                output.write(f" {cat_str}\\{cat_str},\n")
                depth += 1
                indent(output)
                output.write(f"conj({cat_str}\\{cat_str}, {cat_str},")

            if node.op_string == 'conj':
                cat_str = _prolog_category_string(node.cat.left)
                output.write(f" {cat_str},")

            if node.op_string == 'lp':
                cat_str = _prolog_category_string(node.right_child.cat)
                output.write(f" {cat_str},\n")
                depth += 1
                indent(output)
                output.write(f"lp({cat_str},")

            output.write("\n")
            depth += 1
            rec(node.left_child, output)
            if not node.is_unary:
                output.write(",\n")
                rec(node.right_child, output)
            output.write(")")
            depth -= 1

            if node.op_string in ('conj2', 'lp'):
                output.write(")")
                depth -= 1

    with StringIO() as output:
        output.write(f"ccg({sentence_index},\n")
        rec(tree, output)
        output.write(").\n")
        return output.getvalue()


_prolog_header = (
    ':- op(601, xfx, (/)).\n'
    ':- op(601, xfx, (\\)).\n'
    ':- multifile ccg/2, id/2.\n'
    ':- discontiguous ccg/2, id/2.\n'
)


def to_prolog_en(
    nbest_trees: List[List[ScoredTree]],
) -> str:
    """convert parsing results to Prolog format used by LangPro.

    Args:
        nbest_trees (List[List[ScoredTree]]): parsing results

    Returns:
        str: Prolog string
    """

    with StringIO() as output:
        print(_prolog_header, file=output)
        for sentence_index, trees in enumerate(nbest_trees, 1):
            for tree, _ in trees:
                print(_prolog_string(tree, sentence_index), file=output)
        result = output.getvalue()
    return result


_ja_combinators = {
    'SSEQ': 'sseq',
    '>': 'fa',
    '<': 'ba',
    '>B': 'fc',
    '<B1': 'bc1',
    '<B2': 'bc2',
    '<B3': 'bc3',
    '<B4': 'bc4',
    '>Bx1': 'fx1',
    '>Bx2': 'fx2',
    '>Bx3': 'fx3',
    "ADNext": 'adnext',
    "ADNint": 'adnint',
    "ADV0": 'adv0',
    "ADV1": 'adv1',
}


def to_prolog_ja(
    nbest_trees: List[List[ScoredTree]],
) -> str:
    """convert parsing results to Prolog format used by LangPro.
    This is specifically used for Japanese sentences.

    Args:
        nbest_trees (List[List[ScoredTree]]): parsing results

    Returns:
        str: Prolog string
    """

    def traverse_cat(node):
        if node.is_functor:
            left = traverse_cat(node.left)
            right = traverse_cat(node.right)
            return f'({left}{node.slash}{right})'
        else:
            feature = dict(node.feature.items())
            base = node.base.lower()
            if 'case' not in feature:
                return base
            else:
                feature_case = feature["case"].lower()
                return f'{base}:{feature_case}'

    def traverse_tree(node, depth=1):
        whitespace = ' ' * depth
        if node.is_leaf:
            cat = traverse_cat(node.cat)
            token = node.token
            surf = _escape_prolog(token.get('surf', node.word))
            base = _escape_prolog(token.get('base', '*'))

            tags = [
                token.get(key, '*')
                for key in ('pos', 'pos1', 'pos2', 'pos3')
            ]

            if all(tag == '*' for tag in tags):
                pos = '*'
            else:
                pos = '/'.join(_escape_prolog(tag) for tag in tags)

            infl_form = _escape_prolog(token.get('inflectionForm', '*'))
            infl_type = _escape_prolog(token.get('inflectionType', '*'))

            output.write(
                f"\n{whitespace}t({cat}, '{surf}', '{base}', '{pos}', '{infl_form}', '{infl_type}')"
            )
        else:
            cat = traverse_cat(node.cat)
            rule = _ja_combinators[node.op_symbol]
            output.write(f"\n{whitespace}{rule}({cat}")

            for i, child in enumerate(node.children):
                if i < len(node.children):
                    output.write(',')
                traverse_tree(child, depth=depth + 1)

            output.write(')')

    output = StringIO()
    print(_prolog_header, file=output)

    for sentence_index, trees in enumerate(nbest_trees, 1):
        for tree, _ in trees:
            output.write(f'ccg({sentence_index},')
            traverse_tree(tree)
            output.write(').\n\n')

    result = output.getvalue()
    output.close()
    return result



from typing import Dict, List
from overrides import overrides
import json
import logging
import numpy

from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import SequenceLabelField, TextField, MetadataField, ArrayField
from allennlp.data.instance import Instance
from allennlp.data.tokenizers import Token
from allennlp.data.token_indexers import TokenIndexer, SingleIdTokenIndexer

from depccg import utils
from depccg.tools.ja.data import convert_ccgbank_to_json

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def read_dataset_ccgbank_or_json(file_path: str):
    if utils.is_json(file_path):
        logger.info(
            f'Reading instances from lines in json file at: {file_path}'
        )
        with open(file_path, 'r') as data_file:
            json_data = json.load(data_file)
    else:
        logger.info(f'Reading trees in auto file at: {file_path}')
        json_data = convert_ccgbank_to_json(file_path)

    logger.info(f'loaded {len(json_data)} instances')
    return json_data


@DatasetReader.register("ja_supertagging_dataset")
class JaSupertaggingDatasetReader(DatasetReader):
    def __init__(
        self,
        lazy: bool = False,
        token_indexers: Dict[str, TokenIndexer] = None
    ) -> None:
        super().__init__(lazy)
        self._token_indexers = token_indexers or {
            'tokens': SingleIdTokenIndexer()}

    @overrides
    def _read(self, file_path):

        json_data = read_dataset_ccgbank_or_json(cached_path(file_path))
        for instance in json_data:
            sentence, labels = instance
            tags, deps = labels
            yield self.text_to_instance(sentence, tags, deps)

    @overrides
    def text_to_instance(
        self,
        sentence: str,
        tags: List[str] = None,
        deps: List[int] = None,
        weight: float = 1.0
    ) -> Instance:

        tokens = [Token(token) for token in sentence.split(' ')]
        token_field = TextField(tokens, self._token_indexers)
        metadata = MetadataField({'words': sentence})
        weight = ArrayField(numpy.array([weight], 'f'))
        fields = {
            'words': token_field,
            'metadata': metadata,
            'weight': weight,
        }
        if tags is not None and deps is not None:
            fields['head_tags'] = SequenceLabelField(
                tags, token_field, label_namespace='head_tags')
            fields['head_indices'] = SequenceLabelField(
                deps, token_field, label_namespace='head_indices')
        return Instance(fields)



from typing import Dict, List, Optional, Union
from overrides import overrides
import json
import logging
import numpy
import random

from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import SequenceLabelField, TextField, MetadataField, ArrayField
from allennlp.data.instance import Instance
from allennlp.data.tokenizers import Token
from allennlp.data.token_indexers import TokenIndexer, SingleIdTokenIndexer

from depccg import utils
from depccg.tools.data import convert_auto_to_json

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def read_dataset_auto_or_json(file_path: str):
    if utils.is_json(file_path):
        logger.info(
            f'Reading instances from lines in json file at: {file_path}')
        with open(file_path, 'r') as data_file:
            json_data = json.load(data_file)
    else:
        logger.info(f'Reading trees in auto file at: {file_path}')
        json_data = convert_auto_to_json(file_path)
    logger.info(f'loaded {len(json_data)} instances')
    return json_data


@DatasetReader.register("supertagging_dataset")
class SupertaggingDatasetReader(DatasetReader):
    def __init__(
        self,
        lazy: bool = False,
        token_indexers: Dict[str, TokenIndexer] = None
    ) -> None:
        super().__init__(lazy)
        self._token_indexers = token_indexers or {
            'tokens': SingleIdTokenIndexer()}

    @overrides
    def _read(self, file_path):
        json_data = read_dataset_auto_or_json(cached_path(file_path))
        for instance in json_data:
            sentence, labels = instance
            tags, deps = labels
            yield self.text_to_instance(sentence, tags, deps)

    @overrides
    def text_to_instance(
        self,
        sentence: str,
        tags: List[str] = None,
        deps: List[int] = None,
        weight: float = 1.0
    ) -> Instance:  # type: ignore
        # pylint: disable=arguments-differ
        tokens = [
            Token(utils.normalize(token))
            for token in sentence.split(' ')
        ]
        token_field = TextField(tokens, self._token_indexers)
        metadata = MetadataField({'words': sentence})
        weight = ArrayField(numpy.array([weight], 'f'))

        fields = {
            'words': token_field,
            'metadata': metadata,
            'weight': weight,
        }

        if tags is not None and deps is not None:
            fields['head_tags'] = SequenceLabelField(
                tags, token_field, label_namespace='head_tags')
            fields['head_indices'] = SequenceLabelField(
                deps, token_field, label_namespace='head_indices')

        return Instance(fields)


@DatasetReader.register("tritrain_supertagging_dataset")
class TritrainSupertaggingDatasetReader(SupertaggingDatasetReader):
    def __init__(
        self,
        lazy: bool = False,
        noisy_weight: Optional[float] = 0.4,
        token_indexers: Dict[str, TokenIndexer] = None
    ) -> None:
        super().__init__(lazy)
        self.noisy_weight = noisy_weight
        self._token_indexers = token_indexers or {
            'tokens': SingleIdTokenIndexer()}

    @overrides
    def _read(self, file_paths):
        """
        read ccgbank and tritrain datasets (both in json file).
        :param file_paths: a pair of file paths
        :return:
        """
        ccgbank, tritrain = file_paths['ccgbank'], file_paths['tritrain']

        logger.info(f'Reading instances from CCGBank at: {ccgbank}')
        ccgbank = read_dataset_auto_or_json(cached_path(ccgbank))

        logger.info(f'Reading instances from tri-train dataset at: {tritrain}')
        tritrain = read_dataset_auto_or_json(cached_path(tritrain))

        ccgbank_size = len(ccgbank)
        tritrain_size = len(tritrain)
        max_len = ccgbank_size * 15 + tritrain_size
        # make the size of the whole training data equal to that of ccgbank
        # please use this with iterator.cache_instances False!
        for _ in range(ccgbank_size):
            index = random.randint(0, max_len - 1)
            if index < tritrain_size:
                sentence, labels = tritrain[index]
                weight = self.noisy_weight
            else:
                index = (index - tritrain_size) % ccgbank_size
                sentence, labels = ccgbank[index]
                weight = 1.
            tags, deps = labels

            if any(len(word) == 0 for word in sentence.split(' ')):
                logging.info(f'skipping example: {sentence}')
                continue
            new_instance = self.text_to_instance(sentence, tags, deps, weight)
            yield new_instance


def dataset_times(iter, n: Union[int, float]):
    listed = list(iter)
    if isinstance(n, int):
        for _ in range(n):
            random.shuffle(listed)
            for elem in listed:
                yield elem
    else:
        assert isinstance(n, float)
        for _ in range(int(n * len(listed))):
            i = random.randint(0, len(listed) - 1)
            yield listed[i]


@DatasetReader.register("finetune_supertagging_dataset")
class FinetuneSupertaggingDatasetReader(SupertaggingDatasetReader):
    def __init__(
        self,
        lazy: bool = False,
        tritrain_noisy_weight: Optional[float] = 1,
        auxiliary_noisy_weight: Optional[float] = 1,
        ccgbank_ratio: Union[int, float] = 1,
        tritrain_ratio: Union[int, float] = 1,
        auxiliary_ratio: Union[int, float] = 1,
        token_indexers: Dict[str, TokenIndexer] = None
    ) -> None:
        super().__init__(lazy)
        self.tritrain_noisy_weight = tritrain_noisy_weight
        self.auxiliary_noisy_weight = auxiliary_noisy_weight
        self.ccgbank_ratio = ccgbank_ratio
        self.tritrain_ratio = tritrain_ratio
        self.auxiliary_ratio = auxiliary_ratio
        self._token_indexers = token_indexers or {
            'tokens': SingleIdTokenIndexer()}

    @overrides
    def _read(self, file_paths):
        ccgbank, aux_data = file_paths['ccgbank'], file_paths['auxiliary']
        logger.info(f'Reading instances from CCGBank at: {ccgbank}')
        ccgbank = read_dataset_auto_or_json(cached_path(ccgbank))
        logger.info(f'Reading instances from auxiliary dataset at: {aux_data}')
        aux_data = read_dataset_auto_or_json(cached_path(aux_data))

        tritrain = file_paths.get('tritrain', None)
        if tritrain is not None:
            logger.info(
                f'Reading instances from tri-training dataset at: {tritrain}')
            tritrain = read_dataset_auto_or_json(cached_path(tritrain))
        else:
            tritrain = []

        ccgbank = list(dataset_times(ccgbank, self.ccgbank_ratio))
        logger.info(
            f'ccgbank ratio = {self.ccgbank_ratio}: use {len(ccgbank)} sentences from ccgbank'
        )
        tritrain = list(dataset_times(tritrain, self.tritrain_ratio))
        logger.info(
            f'tritrain ratio = {self.tritrain_ratio}: use {len(tritrain)} sentences from ccgbank'
        )
        aux_data = list(dataset_times(aux_data, self.auxiliary_ratio))
        logger.info(
            f'auxiliary ratio = {self.auxiliary_ratio}: use {len(aux_data)} sentences from auxiliary data'
        )
        logger.info(f'noisy weight = {self.tritrain_noisy_weight}')
        logger.info(f'noisy weight = {self.auxiliary_noisy_weight}')

        dataset = (
            [(x, 1.) for x in ccgbank]
            + [(x, self.tritrain_noisy_weight) for x in tritrain]
            + [(x, self.auxiliary_noisy_weight) for x in aux_data]
        )

        dataset_size = len(dataset)
        for _ in range(dataset_size):
            index = random.randint(0, dataset_size - 1)
            (sentence, labels), weight = dataset[index]
            tags, deps = labels
            yield self.text_to_instance(sentence, tags, deps, weight)



from typing import Optional, Callable, Tuple, TypeVar, Iterator
from dataclasses import dataclass
import re

X = TypeVar('X')
Pair = Tuple[X, X]

cat_split = re.compile(r'([\[\]\(\)/\\|])')
punctuations = [',', '.', ';', ':', 'LRB', 'RRB', 'conj', '*START*', '*END*']


class Feature(object):
    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def parse(cls, text: str) -> 'Feature':
        if '=' in text and ',' in text:
            return TernaryFeature(*[tuple(kv.split('=')) for kv in text.split(',')])
        return UnaryFeature(text)


@dataclass(frozen=True, repr=False)
class UnaryFeature(Feature):
    """Common feature type widely used in many CCGBanks.
    This assumes None or "X" values as representing a variable feature.
    As commonly done in the parsing literature, the 'nb' variable is treated
    sometimes as not existing, i.e., NP[conj] and NP[nb] can match.
    """

    value: Optional[str] = None

    def __str__(self) -> str:
        return self.value if self.value is not None else ''

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self == Feature.parse(other)
        elif not isinstance(other, UnaryFeature):
            return False
        return self.value == other.value

    def unifies(self, other: 'UnaryFeature') -> bool:
        return (
            self.is_variable
            or self.is_ignorable
            or self == other
        )

    @property
    def is_variable(self) -> bool:
        return self.value == "X"

    @property
    def is_ignorable(self) -> bool:
        return self.value is None or self.value == "nb"


@dataclass(frozen=True, repr=False)
class TernaryFeature(Feature):
    """Feature type used in the Japanese version of CCGBank.
    This assumes a feature with values (X1, X2, X3) as representing a variable.
    """

    kv1: Pair[str]
    kv2: Pair[str]
    kv3: Pair[str]

    def items(self) -> Iterator[Pair[str]]:
        return (self.kv1, self.kv2, self.kv3)

    def values(self) -> Iterator[str]:
        return (v for _, v in self.items())

    def keys(self) -> Iterator[str]:
        return (k for k, _ in self.items())

    def __str__(self) -> str:
        return ','.join(f'{k}={v}' for k, v in self.items())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self == Feature.parse(other)
        elif not isinstance(other, TernaryFeature):
            return False
        return (
            self.kv1 == other.kv1
            and self.kv2 == other.kv2
            and self.kv3 == other.kv3
        )

    def unifies(self, other: 'UnaryFeature') -> bool:
        if self == other:
            return True

        if list(self.keys()) != list(other.keys()):
            return False

        return all(
            v1 == v2 or v1.startswith('X')
            for v1, v2 in zip(self.values(), other.values())
        )

    @property
    def is_variable(self) -> bool:
        return any(v.startswith('X') for v in self.values())


class Category(object):
    @property
    def is_functor(self):
        return not self.is_atomic

    @property
    def is_atomic(self):
        return not self.is_functor

    def __repr__(self) -> str:
        return str(self)

    def __truediv__(self, other: 'Category') -> 'Category':
        return Functor(self, '/', other)

    def __or__(self, other: 'Category') -> 'Category':
        return Functor(self, '\\', other)

    @classmethod
    def parse(cls, text: str) -> 'Category':
        tokens = cat_split.sub(r' \1 ', text)
        buffer = list(reversed([i for i in tokens.split(' ') if i != '']))
        stack = []

        while len(buffer):
            item = buffer.pop()
            if item in punctuations:
                stack.append(Atom(item))
            elif item == '(':
                pass
            elif item == ')':
                y = stack.pop()
                if len(stack) == 0:
                    return y
                f = stack.pop()
                x = stack.pop()
                stack.append(Functor(x, f, y))
            elif item in ('/', '\\', '|'):
                stack.append(item)
            else:
                if len(buffer) >= 3 and buffer[-1] == '[':
                    buffer.pop()
                    feature = Feature.parse(buffer.pop())
                    assert buffer.pop() == ']'
                    stack.append(Atom(item, feature))
                else:
                    stack.append(Atom(item))

        if len(stack) == 1:
            return stack[0]
        x, f, y = stack
        return Functor(x, f, y)


@dataclass(frozen=True, repr=False)
class Atom(Category):
    base: str
    feature: Feature = UnaryFeature()

    def __str__(self) -> str:
        feature = str(self.feature)
        if len(feature) == 0:
            return self.base
        return f'{self.base}[{feature}]'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return str(self) == other
        elif not isinstance(other, Atom):
            return False
        return (
            self.base == other.base
            and self.feature == other.feature
        )

    def __xor__(self, other: object) -> bool:
        if not isinstance(other, Atom):
            return False
        return self.base == other.base

    @property
    def is_atomic(self):
        return True

    @property
    def nargs(self) -> int:
        return 0

    def arg(self, index: int) -> Optional[Category]:
        if index == 0:
            return self
        return None

    def clear_features(self, *args) -> 'Atom':
        if self.feature in args:
            return Atom(self.base)
        return self


@dataclass(frozen=True, repr=False)
class Functor(Category):
    left: Category
    slash: str
    right: Category

    def __str__(self) -> str:
        def _str(cat):
            if isinstance(cat, Functor):
                return f'({cat})'
            return str(cat)
        return _str(self.left) + self.slash + _str(self.right)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return str(self) == other
        elif not isinstance(other, Functor):
            return False
        return (
            self.left == other.left
            and self.slash == other.slash
            and self.right == other.right
        )

    def __xor__(self, other: object) -> bool:
        if not isinstance(other, Functor):
            return False
        return (
            self.left ^ other.left
            and self.slash == other.slash
            and self.right ^ other.right
        )

    @property
    def functor(self) -> Callable[[Category, Category], Category]:
        return lambda x, y: Functor(x, self.slash, y)

    @property
    def is_functor(self):
        return True

    @property
    def nargs(self) -> int:
        return 1 + self.left.nargs

    def arg(self, index: int) -> Optional[Category]:
        if self.nargs == index:
            return self
        else:
            return self.left.arg(index)

    def clear_features(self, *args) -> 'Functor':
        return self.functor(
            self.left.clear_features(*args),
            self.right.clear_features(*args)
        )



from depccg.tree import Tree
from depccg.utils import normalize


def ja_of(tree: Tree) -> str:
    """tree string in the Japanese CCGBank's format

    Args:
        tree (Tree): tree object

    Returns:
        str: tree string in Japanese CCGBank's format
    """

    def rec(node):
        if node.is_leaf:
            cat = node.cat
            word = normalize(node.word)
            token = node.token

            poss = [
                token.get(pos, '*')
                for pos in ('pos', 'pos1', 'pos2', 'pos3')
            ]

            poss = [pos for pos in poss if pos != '*']
            pos = '-'.join(poss) if len(poss) else '_'

            inflections = [
                token.get(i, '*')
                for i in ('inflectionForm', 'inflectionType')
            ]

            inflections = [i for i in inflections if i != '*']
            inflection = '-'.join(inflections) if len(inflections) else '_'

            return f'{{{cat} {word}/{word}/{pos}/{inflection}}}'
        else:
            children = ' '.join(rec(child) for child in node.children)
            return f'{{{node.op_symbol} {node.cat} {children}}}'

    return rec(tree)



# -*- coding: utf-8 -*-
#
#  Copyright 2015 Pascual Martinez-Gomez
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from lxml import etree

def get_node_at_path(tree, path):
    """
    `tree` is a lxml etree.
    `path` is either a list of indices (integer) or
    a single integer.
    It returns the lxml node of the tree at path `path`.
    """
    assert isinstance(path, list) or isinstance(path, int)
    if isinstance(path, int):
        path = [path]
    node = tree
    for d in path:
        try:
            node = node[d]
        except IndexError:
            raise(IndexError, 'Attempted to index subtree {0} with path {1}'\
                  .format(tree.get('id'), path))
    return node



from typing import List, Optional, Union

import json
from io import StringIO
from lxml import etree

from depccg.tree import ScoredTree
from depccg.instance_models import SEMANTIC_TEMPLATES
from depccg.semantics.ccg2lambda import parse as ccg2lambda
from depccg.lang import get_global_language

from depccg.printer.html import to_mathml
from depccg.printer.jigg_xml import to_jigg_xml
from depccg.printer.prolog import to_prolog_en, to_prolog_ja
from depccg.printer.xml import xml_of
from depccg.printer.ja import ja_of
from depccg.printer.conll import conll_of
from depccg.printer.my_json import json_of
from depccg.printer.deriv import deriv_of
from depccg.printer.ptb import ptb_of
from depccg.printer.auto import auto_of, auto_extended_of


def _process_xml(xml_node):
    return etree \
        .tostring(xml_node, encoding='utf-8', pretty_print=True) \
        .decode('utf-8')


_formatters = {
    'conll': conll_of,
    'auto': auto_of,
    'auto_extended': auto_extended_of,
    'ja': ja_of,
    'deriv': deriv_of,
    'ptb': ptb_of,
}


def to_string(
    nbest_trees: List[Union[List[ScoredTree], ScoredTree]],
    format: str = 'auto',
    semantic_templates: Optional[str] = None,
) -> str:
    """convert parsing results into one string representation

    Args:
        nbest_trees (List[Union[List[ScoredTree], ScoredTree]]): 
            parsed results for multiple sentences
        format (str, optional): format type. Defaults to 'auto'.
        available options are: 'auto', 'auto_extended', 'conll', 'deriv', 'html', 'ja',
        'json', 'ptb', 'jigg_xml', 'jigg_xml_ccg2lambda', 'ccg2lambda', 'prolog'.
        semantic_templates (Optional[str], optional): semantic template used for
        obtaining semantic formula using ccg2lambda. Defaults to None.

    Raises:
        KeyError: if the format option is not supported, this error occurs.

    Returns:
        str: string in the target format
    """
    if isinstance(nbest_trees[0], ScoredTree):
        nbest_trees = [nbest_trees]
    elif not (
        isinstance(nbest_trees[0], list)
        and isinstance(nbest_trees[0][0], ScoredTree)
    ):
        raise RuntimeError('invalid argument type for stringifying trees')

    if format in ('jigg_xml_ccg2lambda', 'ccg2lambda'):
        lang = get_global_language()
        templates = semantic_templates or SEMANTIC_TEMPLATES.get(lang)
        assert templates is not None, \
            f'semantic_templates must be specified for language: {lang}'

    if format == 'conll':
        header = '# ID={}\n# log probability={:.8f}'
    else:
        header = 'ID={}, log probability={:.8f}'

    if format == 'xml':
        return _process_xml(xml_of(nbest_trees))

    elif format == 'jigg_xml':
        return _process_xml(
            to_jigg_xml(
                nbest_trees,
                use_symbol=get_global_language() == 'ja',
            )
        )

    elif format == 'jigg_xml_ccg2lambda':

        jigg_xml = to_jigg_xml(nbest_trees)
        result_xml_str, _ = ccg2lambda.parse(
            jigg_xml, str(templates), ncores=1
        )
        return result_xml_str.decode('utf-8')

    elif format == 'prolog':  # print end=''
        lang = get_global_language()
        if lang == 'en':
            return to_prolog_en(nbest_trees)
        elif lang == 'ja':
            return to_prolog_ja(nbest_trees)
        else:
            raise KeyError(
                f'prolog format is not supported for language {lang}'
            )

    elif format == 'html':
        return to_mathml(nbest_trees)

    elif format == 'json':
        results = {}
        for sentence_index, trees in enumerate(nbest_trees, 1):
            results[sentence_index] = []
            for tree, log_prob in trees:
                tree_dict = json_of(tree)
                tree_dict['log_prob'] = log_prob
                results[sentence_index].append(tree_dict)
        return json.dumps(results, indent=4)

    elif format == 'ccg2lambda':

        with StringIO() as file:
            jigg_xml = to_jigg_xml(nbest_trees)
            _, formulas_list = ccg2lambda.parse(
                jigg_xml, str(templates), ncores=1
            )
            for sentence_index, (trees, formulas) in enumerate(zip(nbest_trees, formulas_list), 1):
                for (tree, log_prob), formula in zip(trees, formulas):
                    print(header.format(sentence_index, log_prob), file=file)
                    print(formula, file=file)
            return file.getvalue()

    try:
        formatter = _formatters[format]
    except KeyError:
        raise KeyError(
            f'unsupported format type: {format}'
        )

    with StringIO() as file:
        for sentence_index, trees in enumerate(nbest_trees, 1):
            for tree, log_prob in trees:
                print(header.format(sentence_index, log_prob), file=file)
                print(formatter(tree), file=file)

        return file.getvalue()


def print_(
    nbest_trees: List[Union[List[ScoredTree], ScoredTree]],
    format: str = 'auto',
    semantic_templates: Optional[str] = None,
    **kwargs,
) -> None:
    """print parsing results into one string representation

    Args:
        nbest_trees (List[Union[List[ScoredTree], ScoredTree]]):
         parsed results for multiple sentences
        format (str, optional): format type. Defaults to 'auto'.
        available options are: 'auto', 'auto_extended', 'conll', 'deriv', 'html', 'ja',
        'json', 'ptb', 'jigg_xml', 'jigg_xml_ccg2lambda', 'ccg2lambda', 'prolog'.
        semantic_templates (Optional[str], optional): semantic template used for
        obtaining semantic formula using ccg2lambda. Defaults to None.

    other keyword arguments for Python 'print' function are also available.

    Raises:
        KeyError: if the format option is not supported, this error occurs.
    """

    print(
        to_string(
            nbest_trees,
            format=format,
            semantic_templates=semantic_templates,
        ),
        **kwargs,
    )
